# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
When given a :class:`GetItem`, the connection will return a :class:`GetItemResponse`:

>>> connection(GetItem(table, {"h": 0}))
<LowVoltage.actions.get_item.GetItemResponse object at ...>

The item is accessed like this:

>>> connection(GetItem(table, {"h": 0})).item
{u'h': 0, u'gr': 0, u'gh': 0}

Note that getting an unexisting item does not raise an exception:

>>> print connection(GetItem(table, {"h": -1})).item
None
"""

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .action import Action
from .conversion import _convert_dict_to_db, _convert_db_to_dict
from .next_gen_mixins import proxy
from .next_gen_mixins import (
    ConsistentRead,
    ExpressionAttributeNames,
    ProjectionExpression,
    ReturnConsumedCapacity,
)
from .return_types import ConsumedCapacity, _is_dict


class GetItemResponse(object):
    """
    The `GetItem response <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_GetItem.html#API_GetItem_ResponseElements>`__
    """

    def __init__(
        self,
        ConsumedCapacity=None,
        Item=None,
        **dummy
    ):
        self.__consumed_capacity = ConsumedCapacity
        self.__item = Item

    @property
    def consumed_capacity(self):
        """
        The capacity consumed by the request. If you used :meth:`~GetItem.return_consumed_capacity_total`.

        :type: ``None`` or :class:`.ConsumedCapacity`
        """
        if _is_dict(self.__consumed_capacity):  # pragma no branch (Defensive code)
            return ConsumedCapacity(**self.__consumed_capacity)

    @property
    def item(self):
        """
        The item you just got. None if the item is not in the table.

        :type: ``None`` or dict
        """
        if _is_dict(self.__item):  # pragma no branch (Defensive code)
            return _convert_db_to_dict(self.__item)


class GetItem(Action):
    """
    The `GetItem request <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_GetItem.html#API_GetItem_RequestParameters>`__
    """

    def __init__(self, table_name, key):
        super(GetItem, self).__init__("GetItem", GetItemResponse)
        self.__table_name = table_name
        self.__key = key
        self.__consistent_read = ConsistentRead(self)
        self.__expression_attribute_names = ExpressionAttributeNames(self)
        self.__projection_expression = ProjectionExpression(self)
        self.__return_consumed_capacity = ReturnConsumedCapacity(self)

    @property
    def payload(self):
        data = {
            "TableName": self.__table_name,
            "Key": _convert_dict_to_db(self.__key),
        }
        data.update(self.__consistent_read.payload)
        data.update(self.__expression_attribute_names.payload)
        data.update(self.__projection_expression.payload)
        data.update(self.__return_consumed_capacity.payload)
        return data

    @proxy
    def consistent_read_true(self):
        """
        >>> connection(
        ...   GetItem(table, {"h": 0})
        ...     .consistent_read_true()
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity.capacity_units
        1.0
        """
        return self.__consistent_read.true()

    @proxy
    def consistent_read_false(self):
        """
        >>> connection(
        ...   GetItem(table, {"h": 0})
        ...     .consistent_read_false()
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity.capacity_units
        0.5
        """
        return self.__consistent_read.false()

    @proxy
    def project(self, *names):
        """
        >>> connection(
        ...   GetItem(table, {"h": 0})
        ...     .project("gr")
        ... ).item
        {u'gr': 0}
        """
        return self.__projection_expression.add(*names)

    @proxy
    def expression_attribute_name(self, synonym, name):
        """
        >>> connection(
        ...   GetItem(table, {"h": 0})
        ...     .expression_attribute_name("syn", "gr")
        ...     .project("#syn")
        ... ).item
        {u'gr': 0}
        """
        return self.__expression_attribute_names.add(synonym, name)

    @proxy
    def return_consumed_capacity_total(self):
        """
        >>> connection(
        ...   GetItem(table, {"h": 0})
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity.capacity_units
        0.5
        """
        return self.__return_consumed_capacity.total()

    @proxy
    def return_consumed_capacity_none(self):
        """
        >>> print connection(
        ...   GetItem(table, {"h": 0})
        ...     .return_consumed_capacity_none()
        ... ).consumed_capacity
        None
        """
        return self.__return_consumed_capacity.none()


class GetItemUnitTests(_tst.UnitTests):
    def test_name(self):
        self.assertEqual(GetItem("Table", {"hash": 42}).name, "GetItem")

    def test_key(self):
        self.assertEqual(
            GetItem("Table", {"hash": 42}).payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"N": "42"}},
            }
        )

    def test_return_consumed_capacity_none(self):
        self.assertEqual(
            GetItem("Table", {"hash": u"h"}).return_consumed_capacity_none().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ReturnConsumedCapacity": "NONE",
            }
        )

    def test_return_consumed_capacity_total(self):
        self.assertEqual(
            GetItem("Table", {"hash": u"h"}).return_consumed_capacity_total().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ReturnConsumedCapacity": "TOTAL",
            }
        )

    def test_consistent_read_true(self):
        self.assertEqual(
            GetItem("Table", {"hash": u"h"}).consistent_read_true().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ConsistentRead": True,
            }
        )

    def test_consistent_read_false(self):
        self.assertEqual(
            GetItem("Table", {"hash": u"h"}).consistent_read_false().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ConsistentRead": False,
            }
        )

    def test_project(self):
        self.assertEqual(
            GetItem("Table", {"hash": u"h"}).project("abc").payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ProjectionExpression": "abc",
            }
        )

    def test_expression_attribute_name(self):
        self.assertEqual(
            GetItem("Table", {"hash": 42}).expression_attribute_name("n", "path").payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"N": "42"}},
                "ExpressionAttributeNames": {"#n": "path"},
            }
        )


class GetItemLocalIntegTests(_tst.LocalIntegTestsWithTableH):
    def test_simple_get(self):
        self.connection(_lv.PutItem("Aaa", {"h": u"get", "a": "yyy"}))

        r = self.connection(_lv.GetItem("Aaa", {"h": u"get"}))

        self.assertEqual(r.item, {"h": "get", "a": "yyy"})

    def test_get_with_projections(self):
        self.connection(_lv.PutItem("Aaa", {"h": u"attrs", "a": "yyy", "b": {"c": ["d1", "d2", "d3"]}, "e": 42, "f": "nope"}))

        r = self.connection(_lv.GetItem("Aaa", {"h": u"attrs"}).project("b.c[1]", "e"))

        self.assertEqual(r.item, {"b": {"c": ["d2"]}, "e": 42})

    def test_unexisting_table(self):
        with self.assertRaises(_lv.ResourceNotFoundException):
            self.connection(_lv.GetItem("Bbb", {}))

    def test_bad_key_type(self):
        with self.assertRaises(_lv.ValidationException):
            self.connection(_lv.GetItem("Aaa", {"h": 42}))


class GetItemConnectedIntegTests(_tst.ConnectedIntegTestsWithTable):
    def setUp(self):
        super(GetItemConnectedIntegTests, self).setUp()
        self.connection(_lv.PutItem(self.table, self.item))

    def tearDown(self):
        self.connection(_lv.DeleteItem(self.table, self.tab_key))
        super(GetItemConnectedIntegTests, self).tearDown()

    def test_return_consumed_capacity_total(self):
        r = self.connection(_lv.GetItem(self.table, self.tab_key).return_consumed_capacity_total())

        self.assertEqual(r.consumed_capacity.capacity_units, 0.5)
        self.assertEqual(r.consumed_capacity.global_secondary_indexes, None)
        self.assertEqual(r.consumed_capacity.local_secondary_indexes, None)
        self.assertEqual(r.consumed_capacity.table, None)
        self.assertEqual(r.consumed_capacity.table_name, self.table)



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import datetime

try:
    from testresources import TestResourceManager, ResourcedTestCase
except ImportError:  # pragma no cover (Test code)
    class TestResourceManager(object):
        pass

    class ResourcedTestCase(object):
        pass

import LowVoltage as _lv


table_name_prefix = datetime.datetime.now().strftime("LowVoltage.Tests.Integ.%Y-%m-%d.%H-%M-%S.%f")


def make_connection():
    return _lv.Connection("eu-west-1", _lv.EnvironmentCredentials())


class ConnectedIntegTests(ResourcedTestCase):
    # Create an IAM user, populate the AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY environment variables
    # Use the following IAM policy to limit access to specific tables:
    # {
    #     "Version": "2012-10-17",
    #     "Statement": [
    #         {
    #             "Action": [
    #                 "dynamodb:*"
    #             ],
    #             "Effect": "Allow",
    #             "Resource": "arn:aws:dynamodb:*:*:table/LowVoltage.Tests.*"
    #         },
    #         {
    #             "Action": [
    #                 "dynamodb:ListTables"
    #             ],
    #             "Effect": "Allow",
    #             "Resource": "*"
    #         }
    #     ]
    # }
    # @todo Could we refine the policy to ensure the total provisioned throughput is capped?

    def setUp(self):
        super(ConnectedIntegTests, self).setUp()
        self.connection = make_connection()

    @classmethod
    def make_table_name(cls):
        assert cls.__name__.endswith("ConnectedIntegTests")
        return "{}.{}".format(table_name_prefix, cls.__name__[:19])


class DynamoDbResourceManager(TestResourceManager):
    def make(self, dependencies):
        connection = make_connection()

        table = table_name_prefix

        connection(
            _lv.CreateTable(table).hash_key("tab_h", _lv.STRING).range_key("tab_r", _lv.NUMBER).provisioned_throughput(1, 1)
                .global_secondary_index("gsi").hash_key("gsi_h", _lv.STRING).range_key("gsi_r", _lv.NUMBER).project_all().provisioned_throughput(1, 1)
                .local_secondary_index("lsi").hash_key("tab_h").range_key("lsi_r", _lv.NUMBER).project_all().provisioned_throughput(1, 1)
        )
        _lv.WaitForTableActivation(connection, table)

        return table

    def clean(self, table):
        connection = make_connection()
        connection(_lv.DeleteTable(table))
        _lv.WaitForTableDeletion(connection, table)


class ConnectedIntegTestsWithTable(ConnectedIntegTests):
    resources = [("table", DynamoDbResourceManager())]

    item = {"tab_h": u"0", "tab_r": 0, "gsi_h": u"1", "gsi_r": 1, "lsi_r": 2}
    tab_key = {"tab_h": u"0", "tab_r": 0}
    gsi_key = {"gsi_h": u"1", "gsi_r": 1}
    lsi_key = {"tab_h": u"0", "lsi_r": 2}



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
When given a :class:`BatchWriteItem`, the connection will return a :class:`BatchWriteItemResponse`:

>>> connection(
...   BatchWriteItem().table(table).delete({"h": 0}, {"h": 1})
... )
<LowVoltage.actions.batch_write_item.BatchWriteItemResponse ...>
"""

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .action import Action
from .conversion import _convert_dict_to_db
from .next_gen_mixins import proxy
from .next_gen_mixins import ReturnConsumedCapacity, ReturnItemCollectionMetrics
from .return_types import ConsumedCapacity, ItemCollectionMetrics, _is_dict, _is_list_of_dict


class BatchWriteItemResponse(object):
    """
    The `BatchWriteItem response <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_BatchWriteItem.html#API_BatchWriteItem_ResponseElements>`__.
    """

    def __init__(
        self,
        ConsumedCapacity=None,
        ItemCollectionMetrics=None,
        UnprocessedItems=None,
        **dummy
    ):
        self.__consumed_capacity = ConsumedCapacity

        self.__item_collection_metrics = ItemCollectionMetrics

        self.__unprocessed_items = UnprocessedItems

    @property
    def consumed_capacity(self):
        """
        The capacity consumed by the request. If you used :meth:`~BatchWriteItem.return_consumed_capacity_total` or :meth:`~BatchWriteItem.return_consumed_capacity_indexes`.

        :type: ``None`` or list of :class:`.ConsumedCapacity`
        """
        if _is_list_of_dict(self.__consumed_capacity):  # pragma no branch (Defensive code)
            return [ConsumedCapacity(**c) for c in self.__consumed_capacity]

    @property
    def item_collection_metrics(self):
        """
        Metrics about the collection of the items you just updated. If a LSI was touched and you used :meth:`~BatchWriteItem.return_item_collection_metrics_size`.

        :type: ``None`` or dict of string (table name) to list of :class:`.ItemCollectionMetrics`
        """
        if _is_dict(self.__item_collection_metrics):  # pragma no branch (Defensive code)
            return {n: [ItemCollectionMetrics(**m) for m in v] for n, v in self.__item_collection_metrics.iteritems()}

    @property
    def unprocessed_items(self):
        """
        Items that were not processed during this request. If not None, you should give this back to the constructor of a subsequent :class:`BatchWriteItem`.

        :type: ``None`` or exactly as returned by DynamoDB
        """
        return self.__unprocessed_items


class BatchWriteItem(Action):
    """
    The `BatchWriteItem request <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_BatchWriteItem.html#API_BatchWriteItem_RequestParameters>`__.
    """

    def __init__(self, previous_unprocessed_items=None):
        super(BatchWriteItem, self).__init__("BatchWriteItem", BatchWriteItemResponse)
        self.__previous_unprocessed_items = previous_unprocessed_items
        self.__tables = {}
        self.__active_table = None
        self.__return_consumed_capacity = ReturnConsumedCapacity(self)
        self.__return_item_collection_metrics = ReturnItemCollectionMetrics(self)

    @property
    def payload(self):
        data = {}
        if self.__previous_unprocessed_items:
            data["RequestItems"] = self.__previous_unprocessed_items
        if self.__tables:
            data["RequestItems"] = {n: t.payload for n, t in self.__tables.iteritems()}
        data.update(self.__return_consumed_capacity.payload)
        data.update(self.__return_item_collection_metrics.payload)
        return data

    class _Table:
        def __init__(self, action):
            self.delete = []
            self.put = []

        @property
        def payload(self):
            items = []
            if self.delete:
                items.extend({"DeleteRequest": {"Key": _convert_dict_to_db(k)}} for k in self.delete)
            if self.put:
                items.extend({"PutRequest": {"Item": _convert_dict_to_db(i)}} for i in self.put)
            return items

    def table(self, name):
        """
        Set the active table. Calls to methods like :meth:`delete` or :meth:`put` will apply to this table.
        """
        if name not in self.__tables:
            self.__tables[name] = self._Table(self)
        self.__active_table = self.__tables[name]
        return self

    def put(self, *items):
        """
        Add items to put in the active table.
        This method accepts a variable number of items or iterables.

        :raise: :exc:`.BuilderError` if called when no table is active.

        >>> connection(
        ...   BatchWriteItem().table(table)
        ...     .put({"h": 11})
        ...     .put({"h": 12}, {"h": 13})
        ...     .put([{"h": 14}, {"h": 15}])
        ...     .put({"h": h} for h in range(16, 20))
        ... )
        <LowVoltage.actions.batch_write_item.BatchWriteItemResponse ...>
        """
        self.__check_active_table()
        for item in items:
            if isinstance(item, dict):
                item = [item]
            self.__active_table.put.extend(item)
        return self

    def delete(self, *keys):
        """
        Add keys to delete from the active table.
        This method accepts a variable number of keys or iterables.

        :raise: :exc:`.BuilderError` if called when no table is active.

        >>> connection(
        ...   BatchWriteItem().table(table)
        ...     .delete({"h": 11})
        ...     .delete({"h": 12}, {"h": 13})
        ...     .delete([{"h": 14}, {"h": 15}])
        ...     .delete({"h": h} for h in range(16, 20))
        ... )
        <LowVoltage.actions.batch_write_item.BatchWriteItemResponse ...>
        """
        self.__check_active_table()
        for key in keys:
            if isinstance(key, dict):
                key = [key]
            self.__active_table.delete.extend(key)
        return self

    @proxy
    def return_consumed_capacity_total(self):
        """
        >>> c = connection(
        ...   BatchWriteItem().table(table).delete({"h": 3})
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity
        >>> c[0].table_name
        u'LowVoltage.Tests.Doc.1'
        >>> c[0].capacity_units
        2.0
        """
        return self.__return_consumed_capacity.total()

    @proxy
    def return_consumed_capacity_indexes(self):
        """
        >>> c = connection(
        ...   BatchWriteItem().table(table).delete({"h": 4})
        ...     .return_consumed_capacity_indexes()
        ... ).consumed_capacity
        >>> c[0].capacity_units
        2.0
        >>> c[0].table.capacity_units
        1.0
        >>> c[0].global_secondary_indexes["gsi"].capacity_units
        1.0
        """
        return self.__return_consumed_capacity.indexes()

    @proxy
    def return_consumed_capacity_none(self):
        """
        >>> print connection(
        ...   BatchWriteItem().table(table).delete({"h": 5})
        ...     .return_consumed_capacity_none()
        ... ).consumed_capacity
        None
        """
        return self.__return_consumed_capacity.none()

    @proxy
    def return_item_collection_metrics_size(self):
        """
        >>> m = connection(
        ...   BatchWriteItem().table(table2).put({"h": 0, "r1": 0, "r2": 0})
        ...     .return_item_collection_metrics_size()
        ... ).item_collection_metrics
        >>> m[table2][0].item_collection_key
        {u'h': 0}
        >>> m[table2][0].size_estimate_range_gb
        [0.0, 1.0]
        """
        return self.__return_item_collection_metrics.size()

    @proxy
    def return_item_collection_metrics_none(self):
        """
        >>> print connection(
        ...   BatchWriteItem().table(table2).put({"h": 1, "r1": 0, "r2": 0})
        ...     .return_item_collection_metrics_none()
        ... ).item_collection_metrics
        None
        """
        return self.__return_item_collection_metrics.none()

    def __check_active_table(self):
        if self.__active_table is None:
            raise _lv.BuilderError("No active table.")


class BatchWriteItemUnitTests(_tst.UnitTests):
    def test_name(self):
        self.assertEqual(BatchWriteItem().name, "BatchWriteItem")

    def test_empty(self):
        self.assertEqual(
            BatchWriteItem().payload,
            {}
        )

    def test_return_consumed_capacity_none(self):
        self.assertEqual(
            BatchWriteItem().return_consumed_capacity_none().payload,
            {
                "ReturnConsumedCapacity": "NONE",
            }
        )

    def test_return_consumed_capacity_indexes(self):
        self.assertEqual(
            BatchWriteItem().return_consumed_capacity_indexes().payload,
            {
                "ReturnConsumedCapacity": "INDEXES",
            }
        )

    def test_return_consumed_capacity_total(self):
        self.assertEqual(
            BatchWriteItem().return_consumed_capacity_total().payload,
            {
                "ReturnConsumedCapacity": "TOTAL",
            }
        )

    def test_return_item_collection_metrics_none(self):
        self.assertEqual(
            BatchWriteItem().return_item_collection_metrics_none().payload,
            {
                "ReturnItemCollectionMetrics": "NONE",
            }
        )

    def test_return_item_collection_metrics_size(self):
        self.assertEqual(
            BatchWriteItem().return_item_collection_metrics_size().payload,
            {
                "ReturnItemCollectionMetrics": "SIZE",
            }
        )

    def test_table(self):
        self.assertEqual(
            BatchWriteItem().table("Table").payload,
            {
                "RequestItems": {
                    "Table": [
                    ],
                },
            }
        )

    def test_delete(self):
        self.assertEqual(
            BatchWriteItem().table("Table").delete({"hash": u"h1"}).table("Table").delete([{"hash": u"h2"}]).payload,
            {
                "RequestItems": {
                    "Table": [
                        {"DeleteRequest": {"Key": {"hash": {"S": "h1"}}}},
                        {"DeleteRequest": {"Key": {"hash": {"S": "h2"}}}},
                    ],
                },
            }
        )

    def test_put(self):
        self.assertEqual(
            BatchWriteItem().table("Table").put({"hash": u"h1"}, [{"hash": u"h2"}]).payload,
            {
                "RequestItems": {
                    "Table": [
                        {"PutRequest": {"Item": {"hash": {"S": "h1"}}}},
                        {"PutRequest": {"Item": {"hash": {"S": "h2"}}}},
                    ],
                },
            }
        )

    def test_alternate_between_tables_and_put_delete(self):
        self.assertEqual(
            BatchWriteItem()
                .table("Table1").delete({"hash": u"h1"})
                .table("Table2").put([{"hash": u"h2"}])
                .table("Table1").put({"hash": u"h11"})
                .table("Table2").delete({"hash": u"h22"})
                .payload,
            {
                "RequestItems": {
                    "Table1": [
                        {"DeleteRequest": {"Key": {"hash": {"S": "h1"}}}},
                        {"PutRequest": {"Item": {"hash": {"S": "h11"}}}},
                    ],
                    "Table2": [
                        {"DeleteRequest": {"Key": {"hash": {"S": "h22"}}}},
                        {"PutRequest": {"Item": {"hash": {"S": "h2"}}}},
                    ],
                },
            }
        )

    def test_put_without_active_table(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            BatchWriteItem().put({"h": 0})
        self.assertEqual(catcher.exception.args, ("No active table.",))

    def test_delete_without_active_table(self):
        with self.assertRaises(_lv.BuilderError) as catcher:
            BatchWriteItem().delete({"h": 0})
        self.assertEqual(catcher.exception.args, ("No active table.",))


class BatchWriteItemLocalIntegTests(_tst.LocalIntegTestsWithTableH):
    def test_simple_batch_put(self):
        r = self.connection(_lv.BatchWriteItem().table("Aaa").put(
            {"h": u"1", "a": "xxx"},
            {"h": u"2", "a": "yyy"},
            {"h": u"3", "a": "zzz"},
        ))

        self.assertEqual(self.connection(_lv.GetItem("Aaa", {"h": u"1"})).item, {"h": "1", "a": "xxx"})
        self.assertEqual(self.connection(_lv.GetItem("Aaa", {"h": u"2"})).item, {"h": "2", "a": "yyy"})
        self.assertEqual(self.connection(_lv.GetItem("Aaa", {"h": u"3"})).item, {"h": "3", "a": "zzz"})

    def test_simple_batch_delete(self):
        self.connection(_lv.BatchWriteItem().table("Aaa").put(
            {"h": u"1", "a": "xxx"},
            {"h": u"2", "a": "yyy"},
            {"h": u"3", "a": "zzz"},
        ))

        r = self.connection(_lv.BatchWriteItem().table("Aaa").delete(
            {"h": u"1"},
            {"h": u"2"},
            {"h": u"3"}
        ))

        self.assertEqual(self.connection(_lv.GetItem("Aaa", {"h": u"1"})).item, None)
        self.assertEqual(self.connection(_lv.GetItem("Aaa", {"h": u"2"})).item, None)
        self.assertEqual(self.connection(_lv.GetItem("Aaa", {"h": u"3"})).item, None)


class BatchWriteItemConnectedIntegTests(_tst.ConnectedIntegTestsWithTable):
    def tearDown(self):
        self.connection(_lv.DeleteItem(self.table, self.tab_key))
        super(BatchWriteItemConnectedIntegTests, self).tearDown()

    def test_return_consumed_capacity_indexes(self):
        r = self.connection(_lv.BatchWriteItem().table(self.table).put(self.item).return_consumed_capacity_indexes())

        self.assertEqual(r.consumed_capacity[0].capacity_units, 3.0)
        self.assertEqual(r.consumed_capacity[0].global_secondary_indexes["gsi"].capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity[0].local_secondary_indexes["lsi"].capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity[0].table.capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity[0].table_name, self.table)

    def test_return_item_collection_metrics_size(self):
        r = self.connection(_lv.BatchWriteItem().table(self.table).put(self.item).return_item_collection_metrics_size())

        self.assertEqual(r.item_collection_metrics[self.table][0].item_collection_key, {"tab_h": "0"})
        self.assertEqual(r.item_collection_metrics[self.table][0].size_estimate_range_gb[0], 0.0)
        self.assertEqual(r.item_collection_metrics[self.table][0].size_estimate_range_gb[1], 1.0)



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>


class Iterator(object):
    # Don't implement anything else than a single forward iteration.
    # Remember PyGithub's PaginatedList; it was too difficult to maintain for niche use-cases.
    # Clients can use raw actions to implement their specific needs.

    def __init__(self, connection, first_action):
        self.__connection = connection
        self.__current_iter = [].__iter__()
        self.__next_action = first_action

    def __iter__(self):
        return self

    def next(self):
        try:
            return self.__current_iter.next()
        except StopIteration:
            if self.__next_action is None:
                raise
            else:
                r = self.__connection(self.__next_action)
                next_action, new_items = self.process(self.__next_action, r)
                self.__next_action = next_action
                self.__current_iter = new_items.__iter__()
                return self.__current_iter.next()



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
All public symbols are visible in the top-level module, no need to import submodules.
"""

from actions import *
from attribute_types import *
from compounds import *
from connection import *
from exceptions import *

# @todo __str__ and __repr__
# @todo create builder for attribute paths
# @todo improve builder for expressions
# @todo metrics
# @todo debug logging



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import time

import LowVoltage as _lv
import LowVoltage.testing as _tst


# @todo Should we abort if the table_status is not "DELETING"?
def WaitForTableDeletion(connection, table):
    """Make "DescribeTable" actions until a ResourceNotFoundException is raised."""

    while True:
        try:
            connection(_lv.DescribeTable(table))
            time.sleep(3)
        except _lv.ResourceNotFoundException:
            break


class WaitForTableDeletionLocalIntegTests(_tst.LocalIntegTests):
    def setUp(self):
        super(WaitForTableDeletionLocalIntegTests, self).setUp()
        self.connection(_lv.CreateTable("Aaa").hash_key("h", _lv.STRING).provisioned_throughput(1, 1))

    def test(self):
        self.connection(_lv.DeleteTable("Aaa"))
        _lv.WaitForTableDeletion(self.connection, "Aaa")
        with self.assertRaises(_lv.ResourceNotFoundException):
            self.connection(_lv.DescribeTable("Aaa"))


class WaitForTableDeletionConnectedIntegTests(_tst.ConnectedIntegTests):
    def setUp(self):
        super(WaitForTableDeletionConnectedIntegTests, self).setUp()
        self.table = self.make_table_name()
        self.connection(_lv.CreateTable(self.table).hash_key("h", _lv.STRING).provisioned_throughput(1, 1))
        _lv.WaitForTableActivation(self.connection, self.table)

    def test(self):
        self.connection(_lv.DeleteTable(self.table))
        _lv.WaitForTableDeletion(self.connection, self.table)
        with self.assertRaises(_lv.ResourceNotFoundException):
            self.connection(_lv.DescribeTable(self.table))



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>
"""
When given a :class:`PutItem`, the connection will return a :class:`PutItemResponse`:

>>> connection(PutItem(table, {"h": 0}))
<LowVoltage.actions.put_item.PutItemResponse object at ...>
"""

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .action import Action
from .conversion import _convert_dict_to_db, _convert_db_to_dict
from .next_gen_mixins import proxy
from .next_gen_mixins import (
    ConditionExpression,
    ExpressionAttributeNames,
    ExpressionAttributeValues,
    ReturnConsumedCapacity,
    ReturnItemCollectionMetrics,
    ReturnValues,
)
from .return_types import ItemCollectionMetrics, ConsumedCapacity, _is_dict


class PutItemResponse(object):
    """
    The `PutItem response <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_PutItem.html#API_PutItem_ResponseElements>`__.
    """

    def __init__(
        self,
        Attributes=None,
        ConsumedCapacity=None,
        ItemCollectionMetrics=None,
        **dummy
    ):
        self.__attributes = Attributes
        self.__consumed_capacity = ConsumedCapacity
        self.__item_collection_metrics = ItemCollectionMetrics

    @property
    def attributes(self):
        """
        The previous attributes of the item you just put. If you used :meth:`~PutItem.return_values_all_old`.

        :type: ``None`` or dict
        """
        if _is_dict(self.__attributes):  # pragma no branch (Defensive code)
            return _convert_db_to_dict(self.__attributes)

    @property
    def consumed_capacity(self):
        """
        The capacity consumed by the request. If you used :meth:`~PutItem.return_consumed_capacity_total` or :meth:`~PutItem.return_consumed_capacity_indexes`.

        :type: ``None`` or :class:`.ConsumedCapacity`
        """
        if _is_dict(self.__consumed_capacity):  # pragma no branch (Defensive code)
            return ConsumedCapacity(**self.__consumed_capacity)

    @property
    def item_collection_metrics(self):
        """
        Metrics about the collection of the item you just put. If a LSI was touched and you used :meth:`~PutItem.return_item_collection_metrics_size`.

        :type: ``None`` or :class:`.ItemCollectionMetrics`
        """
        if _is_dict(self.__item_collection_metrics):  # pragma no branch (Defensive code)
            return ItemCollectionMetrics(**self.__item_collection_metrics)


class PutItem(Action):
    """
    The `PutItem request <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_PutItem.html#API_PutItem_RequestParameters>`__.
    """

    def __init__(self, table_name, item):
        super(PutItem, self).__init__("PutItem", PutItemResponse)
        self.__table_name = table_name
        self.__item = item
        self.__condition_expression = ConditionExpression(self)
        self.__expression_attribute_names = ExpressionAttributeNames(self)
        self.__expression_attribute_values = ExpressionAttributeValues(self)
        self.__return_consumed_capacity = ReturnConsumedCapacity(self)
        self.__return_item_collection_metrics = ReturnItemCollectionMetrics(self)
        self.__return_values = ReturnValues(self)

    @property
    def payload(self):
        data = {
            "TableName": self.__table_name,
            "Item": _convert_dict_to_db(self.__item),
        }
        data.update(self.__condition_expression.payload)
        data.update(self.__expression_attribute_names.payload)
        data.update(self.__expression_attribute_values.payload)
        data.update(self.__return_consumed_capacity.payload)
        data.update(self.__return_item_collection_metrics.payload)
        data.update(self.__return_values.payload)
        return data

    @proxy
    def condition_expression(self, expression):
        """
        >>> connection(
        ...   PutItem(table, {"h": 1})
        ...     .condition_expression("#syn=:val")
        ...     .expression_attribute_name("syn", "gr")
        ...     .expression_attribute_value("val", 0)
        ... )
        <LowVoltage.actions.put_item.PutItemResponse object at ...>
        """
        return self.__condition_expression.set(expression)

    @proxy
    def expression_attribute_name(self, synonym, name):
        """
        See :meth:`condition_expression` for an example.
        """
        return self.__expression_attribute_names.add(synonym, name)

    @proxy
    def expression_attribute_value(self, name, value):
        """
        See :meth:`condition_expression` for an example.
        """
        return self.__expression_attribute_values.add(name, value)

    @proxy
    def return_values_all_old(self):
        """
        >>> connection(
        ...   PutItem(table, {"h": 2})
        ...     .return_values_all_old()
        ... ).attributes
        {u'h': 2, u'gr': 0, u'gh': 0}
        """
        return self.__return_values.all_old()

    @proxy
    def return_values_none(self):
        """
        >>> print connection(
        ...   PutItem(table, {"h": 3})
        ...     .return_values_none()
        ... ).attributes
        None
        """
        return self.__return_values.none()

    @proxy
    def return_consumed_capacity_total(self):
        """
        >>> connection(
        ...   PutItem(table, {"h": 4, "gh": 4, "gr": 4})
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity.capacity_units
        3.0
        """
        return self.__return_consumed_capacity.total()

    @proxy
    def return_consumed_capacity_indexes(self):
        """
        >>> c = connection(
        ...   PutItem(table, {"h": 5, "gh": 5, "gr": 5})
        ...     .return_consumed_capacity_indexes()
        ... ).consumed_capacity
        >>> c.capacity_units
        3.0
        >>> c.table.capacity_units
        1.0
        >>> c.global_secondary_indexes["gsi"].capacity_units
        2.0
        """
        return self.__return_consumed_capacity.indexes()

    @proxy
    def return_consumed_capacity_none(self):
        """
        >>> print connection(
        ...   PutItem(table, {"h": 6})
        ...     .return_consumed_capacity_none()
        ... ).consumed_capacity
        None
        """
        return self.__return_consumed_capacity.none()

    @proxy
    def return_item_collection_metrics_size(self):
        """
        >>> m = connection(
        ...   PutItem(table2, {"h": 0, "r1": 0, "r2": 1})
        ...     .return_item_collection_metrics_size()
        ... ).item_collection_metrics
        >>> m.item_collection_key
        {u'h': 0}
        >>> m.size_estimate_range_gb
        [0.0, 1.0]
        """
        return self.__return_item_collection_metrics.size()

    @proxy
    def return_item_collection_metrics_none(self):
        """
        >>> print connection(
        ...   PutItem(table2, {"h": 1, "r1": 0, "r2": 1})
        ...     .return_item_collection_metrics_none()
        ... ).item_collection_metrics
        None
        """
        return self.__return_item_collection_metrics.none()


class PutItemUnitTests(_tst.UnitTests):
    def test_name(self):
        self.assertEqual(PutItem("Table", {"hash": 42}).name, "PutItem")

    def test_item(self):
        self.assertEqual(
            PutItem("Table", {"hash": u"value"}).payload,
            {
                "TableName": "Table",
                "Item": {"hash": {"S": "value"}},
            }
        )

    def test_return_values_none(self):
        self.assertEqual(
            PutItem("Table", {"hash": u"h"}).return_values_none().payload,
            {
                "TableName": "Table",
                "Item": {"hash": {"S": "h"}},
                "ReturnValues": "NONE",
            }
        )

    def test_return_values_all_old(self):
        self.assertEqual(
            PutItem("Table", {"hash": u"h"}).return_values_all_old().payload,
            {
                "TableName": "Table",
                "Item": {"hash": {"S": "h"}},
                "ReturnValues": "ALL_OLD",
            }
        )

    def test_return_consumed_capacity_total(self):
        self.assertEqual(
            PutItem("Table", {"hash": u"h"}).return_consumed_capacity_total().payload,
            {
                "TableName": "Table",
                "Item": {"hash": {"S": "h"}},
                "ReturnConsumedCapacity": "TOTAL",
            }
        )

    def test_return_consumed_capacity_indexes(self):
        self.assertEqual(
            PutItem("Table", {"hash": u"h"}).return_consumed_capacity_indexes().payload,
            {
                "TableName": "Table",
                "Item": {"hash": {"S": "h"}},
                "ReturnConsumedCapacity": "INDEXES",
            }
        )

    def test_return_consumed_capacity_none(self):
        self.assertEqual(
            PutItem("Table", {"hash": u"h"}).return_consumed_capacity_none().payload,
            {
                "TableName": "Table",
                "Item": {"hash": {"S": "h"}},
                "ReturnConsumedCapacity": "NONE",
            }
        )

    def test_return_item_collection_metrics_size(self):
        self.assertEqual(
            PutItem("Table", {"hash": u"h"}).return_item_collection_metrics_size().payload,
            {
                "TableName": "Table",
                "Item": {"hash": {"S": "h"}},
                "ReturnItemCollectionMetrics": "SIZE",
            }
        )

    def test_return_item_collection_metrics_none(self):
        self.assertEqual(
            PutItem("Table", {"hash": u"h"}).return_item_collection_metrics_none().payload,
            {
                "TableName": "Table",
                "Item": {"hash": {"S": "h"}},
                "ReturnItemCollectionMetrics": "NONE",
            }
        )

    def test_expression_attribute_value(self):
        self.assertEqual(
            PutItem("Table", {"hash": 42}).expression_attribute_value("v", u"value").payload,
            {
                "TableName": "Table",
                "Item": {"hash": {"N": "42"}},
                "ExpressionAttributeValues": {":v": {"S": "value"}},
            }
        )

    def test_expression_attribute_name(self):
        self.assertEqual(
            PutItem("Table", {"hash": 42}).expression_attribute_name("n", "path").payload,
            {
                "TableName": "Table",
                "Item": {"hash": {"N": "42"}},
                "ExpressionAttributeNames": {"#n": "path"},
            }
        )

    def test_condition_expression(self):
        self.assertEqual(
            PutItem("Table", {"hash": 42}).condition_expression("a=b").payload,
            {
                "TableName": "Table",
                "Item": {"hash": {"N": "42"}},
                "ConditionExpression": "a=b",
            }
        )


class PutItemLocalIntegTests(_tst.LocalIntegTestsWithTableH):
    def test_simple_put(self):
        self.connection(_lv.PutItem("Aaa", {"h": u"simple"}))

        self.assertEqual(self.connection(_lv.GetItem("Aaa", {"h": u"simple"})).item, {"h": u"simple"})

    def test_put_all_types(self):
        self.connection(_lv.PutItem("Aaa", {
            "h": u"all",
            "number": 42,
            "string": u"àoé",
            "binary": b"\xFF\x00\xFF",
            "bool 1": True,
            "bool 2": False,
            "null": None,
            "number set": set([42, 43]),
            "string set": set([u"éoà", u"bar"]),
            "binary set": set([b"\xFF", b"\xAB"]),
            "list": [True, 42],
            "map": {"a": True, "b": 42},
        }))

        self.assertEqual(
            self.connection(_lv.GetItem("Aaa", {"h": u"all"})).item,
            {
                "h": u"all",
                "number": 42,
                "string": u"àoé",
                "binary": b"\xFF\x00\xFF",
                "bool 1": True,
                "bool 2": False,
                "null": None,
                "number set": set([42, 43]),
                "string set": set([u"éoà", u"bar"]),
                "binary set": set([b"\xFF", b"\xAB"]),
                "list": [True, 42],
                "map": {"a": True, "b": 42},
            }
        )

    def test_return_old_values(self):
        self.connection(PutItem("Aaa", {"h": u"return", "a": b"yyy"}))

        r = self.connection(
            PutItem("Aaa", {"h": u"return", "b": b"xxx"}).return_values_all_old()
        )

        self.assertEqual(r.attributes, {"h": u"return", "a": b"yyy"})


class PutItemConnectedIntegTests(_tst.ConnectedIntegTestsWithTable):
    def tearDown(self):
        self.connection(_lv.DeleteItem(self.table, self.tab_key))
        super(PutItemConnectedIntegTests, self).tearDown()

    def test_return_consumed_capacity_indexes(self):
        r = self.connection(_lv.PutItem(self.table, self.item).return_consumed_capacity_indexes())

        self.assertEqual(r.consumed_capacity.capacity_units, 3.0)
        self.assertEqual(r.consumed_capacity.global_secondary_indexes["gsi"].capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity.local_secondary_indexes["lsi"].capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity.table.capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity.table_name, self.table)

    def test_return_item_collection_metrics_size(self):
        r = self.connection(_lv.PutItem(self.table, self.item).return_item_collection_metrics_size())

        self.assertEqual(r.item_collection_metrics.item_collection_key, {"tab_h": "0"})
        self.assertEqual(r.item_collection_metrics.size_estimate_range_gb[0], 0.0)
        self.assertEqual(r.item_collection_metrics.size_estimate_range_gb[1], 1.0)



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
When given a :class:`UpdateItem`, the connection will return a :class:`UpdateItemResponse`:

>>> connection(UpdateItem(table, {"h": 0}).remove("a"))
<LowVoltage.actions.update_item.UpdateItemResponse object at ...>
"""

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .action import Action
from .conversion import _convert_dict_to_db, _convert_db_to_dict
from .next_gen_mixins import proxy
from .next_gen_mixins import (
    ConditionExpression,
    ExpressionAttributeNames,
    ExpressionAttributeValues,
    ReturnConsumedCapacity,
    ReturnItemCollectionMetrics,
    ReturnValues,
)
from .return_types import ConsumedCapacity, ItemCollectionMetrics, _is_dict


class UpdateItemResponse(object):
    """
    The `UpdateItem response <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_UpdateItem.html#API_UpdateItem_ResponseElements>`__.
    """

    def __init__(
        self,
        Attributes=None,
        ConsumedCapacity=None,
        ItemCollectionMetrics=None,
        **dummy
    ):
        self.__attributes = Attributes
        self.__consumed_capacity = ConsumedCapacity
        self.__item_collection_metrics = ItemCollectionMetrics

    @property
    def attributes(self):
        """
        The (previous or new) attributes of the item you just updated. If you used :meth:`~UpdateItem.return_values_all_old`, :meth:`~UpdateItem.return_values_all_new`, :meth:`~UpdateItem.return_values_updated_old` or :meth:`~UpdateItem.return_values_updated_new`.

        :type: ``None`` or dict
        """
        if _is_dict(self.__attributes):  # pragma no branch (Defensive code)
            return _convert_db_to_dict(self.__attributes)

    # @todo Remove '_' prefix from classes in .return_types.

    @property
    def consumed_capacity(self):
        """
        The capacity consumed by the request. If you used :meth:`~UpdateItem.return_consumed_capacity_total` or :meth:`~UpdateItem.return_consumed_capacity_indexes`.

        :type: ``None`` or :class:`.ConsumedCapacity`
        """
        if _is_dict(self.__consumed_capacity):  # pragma no branch (Defensive code)
            return ConsumedCapacity(**self.__consumed_capacity)

    @property
    def item_collection_metrics(self):
        """
        Metrics about the collection of the item you just updated. If a LSI was touched and you used :meth:`~UpdateItem.return_item_collection_metrics_size`.

        :type: ``None`` or :class:`.ItemCollectionMetrics`
        """
        if _is_dict(self.__item_collection_metrics):  # pragma no branch (Defensive code)
            return ItemCollectionMetrics(**self.__item_collection_metrics)


class UpdateItem(Action):
    """
    The `UpdateItem request <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_UpdateItem.html#API_UpdateItem_RequestParameters>`__.
    """

    def __init__(self, table_name, key):
        super(UpdateItem, self).__init__("UpdateItem", UpdateItemResponse)
        self.__table_name = table_name
        self.__key = key
        self.__set = {}
        self.__remove = []
        self.__add = {}
        self.__delete = {}
        self.__condition_expression = ConditionExpression(self)
        self.__expression_attribute_names = ExpressionAttributeNames(self)
        self.__expression_attribute_values = ExpressionAttributeValues(self)
        self.__return_consumed_capacity = ReturnConsumedCapacity(self)
        self.__return_item_collection_metrics = ReturnItemCollectionMetrics(self)
        self.__return_values = ReturnValues(self)

    @property
    def payload(self):
        data = {
            "TableName": self.__table_name,
            "Key": _convert_dict_to_db(self.__key),
        }
        update = []
        if self.__set:
            update.append("SET {}".format(", ".join("{}={}".format(n, v) for n, v in self.__set.iteritems())))
        if self.__remove:
            update.append("REMOVE {}".format(", ".join(self.__remove)))
        if self.__add:
            update.append("ADD {}".format(", ".join("{} :{}".format(n, v) for n, v in self.__add.iteritems())))
        if self.__delete:
            update.append("DELETE {}".format(", ".join("{} :{}".format(n, v) for n, v in self.__delete.iteritems())))
        if update:
            data["UpdateExpression"] = " ".join(update)
        data.update(self.__condition_expression.payload)
        data.update(self.__expression_attribute_names.payload)
        data.update(self.__expression_attribute_values.payload)
        data.update(self.__return_consumed_capacity.payload)
        data.update(self.__return_item_collection_metrics.payload)
        data.update(self.__return_values.payload)
        return data

    # @todo should we provide bundle methods for set, add, delete that do an implicit expression_attribute_value with a generated value_name?
    # @todo should we provide add_to_int (accepting an int), add_to_set and delete_from_set (accepting several ints, strs or binaries)?

    def set(self, attribute_name, value_name):
        """
        Add a value to SET as an attribute to UpdateExpression.
        As described in the `developer guide <http://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Expressions.Modifying.html#Expressions.Modifying.UpdateExpressions.SET>`__.

        >>> connection(PutItem(table, {"h": 0}))
        <LowVoltage.actions.put_item.PutItemResponse object at ...>
        >>> connection(
        ...   UpdateItem(table, {"h": 0})
        ...     .set("a", ":forty_two")
        ...     .expression_attribute_value("forty_two", 42)
        ...     .return_values_all_new()
        ... ).attributes
        {u'a': 42, u'h': 0}
        """
        self.__set[attribute_name] = value_name
        return self

    def remove(self, path):
        """
        Add an attribute to REMOVE to UpdateExpression.
        As described in the `developer guide <http://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Expressions.Modifying.html#Expressions.Modifying.UpdateExpressions.REMOVE>`__.

        >>> connection(PutItem(table, {"h": 0, "a": 42}))
        <LowVoltage.actions.put_item.PutItemResponse object at ...>
        >>> connection(
        ...   UpdateItem(table, {"h": 0})
        ...     .remove("a")
        ...     .return_values_all_new()
        ... ).attributes
        {u'h': 0}
        """
        self.__remove.append(path)
        return self

    def add(self, attribute_name, value_name):
        """
        Add a (set of) value(s) to ADD to a number (or a set) attribute to UpdateExpression.
        As described in the `developer guide <http://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Expressions.Modifying.html#Expressions.Modifying.UpdateExpressions.ADD>`__.

        >>> connection(PutItem(table, {"h": 0, "a": 42}))
        <LowVoltage.actions.put_item.PutItemResponse object at ...>
        >>> connection(
        ...   UpdateItem(table, {"h": 0})
        ...     .add("a", "two")
        ...     .expression_attribute_value("two", 2)
        ...     .return_values_all_new()
        ... ).attributes
        {u'a': 44, u'h': 0}

        >>> connection(PutItem(table, {"h": 0, "a": {2, 3}}))
        <LowVoltage.actions.put_item.PutItemResponse object at ...>
        >>> connection(
        ...   UpdateItem(table, {"h": 0})
        ...     .add("a", "vals")
        ...     .expression_attribute_value("vals", {1, 2})
        ...     .return_values_all_new()
        ... ).attributes
        {u'a': set([1, 2, 3]), u'h': 0}
        """
        self.__add[attribute_name] = value_name
        return self

    # @todo What happens if you add and delete from the same set? The same values?
    # @todo What happens if you add twice to the same number?
    # @todo What happens if you add twice to the same set?
    # @todo What happens if you delete twice from the same set?
    # @todo What happens if you set the same attribute twice?
    # @todo What happens if you remove the same attribute twice?
    # @todo What happens if you set and remove the same attribute?
    # @todo Should we handle those cases in a "the last call wins" manner like we do for return_values_xxx?

    def delete(self, attribute_name, value_name):
        """
        Add a set of values to DELETE from a set attribute to UpdateExpression.
        As described in the `developer guide <http://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Expressions.Modifying.html#Expressions.Modifying.UpdateExpressions.DELETE>`__.

        >>> connection(PutItem(table, {"h": 0, "a": {1, 2, 3}}))
        <LowVoltage.actions.put_item.PutItemResponse object at ...>
        >>> connection(
        ...   UpdateItem(table, {"h": 0})
        ...     .delete("a", "vals")
        ...     .expression_attribute_value("vals", {1, 2, 4})
        ...     .return_values_all_new()
        ... ).attributes
        {u'a': set([3]), u'h': 0}
        """
        self.__delete[attribute_name] = value_name
        return self

    @proxy
    def condition_expression(self, expression):
        """
        >>> connection(
        ...   UpdateItem(table, {"h": 1})
        ...     .remove("gh")
        ...     .condition_expression("#syn=:val")
        ...     .expression_attribute_name("syn", "gr")
        ...     .expression_attribute_value("val", 0)
        ... )
        <LowVoltage.actions.update_item.UpdateItemResponse object at ...>
        """
        return self.__condition_expression.set(expression)

    @proxy
    def expression_attribute_name(self, synonym, name):
        """
        See :meth:`condition_expression` for an example.
        """
        return self.__expression_attribute_names.add(synonym, name)

    @proxy
    def expression_attribute_value(self, name, value):
        """
        See :meth:`condition_expression` for an example.
        """
        return self.__expression_attribute_values.add(name, value)

    @proxy
    def return_values_all_old(self):
        """
        >>> connection(PutItem(table, {"h": 0, "a": 1, "b": 2}))
        <LowVoltage.actions.put_item.PutItemResponse object at ...>
        >>> connection(
        ...   UpdateItem(table, {"h": 0})
        ...     .set("a", ":v")
        ...     .expression_attribute_value("v", 2)
        ...     .return_values_all_old()
        ... ).attributes
        {u'a': 1, u'h': 0, u'b': 2}
        """
        return self.__return_values.all_old()

    @proxy
    def return_values_all_new(self):
        """
        >>> connection(PutItem(table, {"h": 0, "a": 1, "b": 2}))
        <LowVoltage.actions.put_item.PutItemResponse object at ...>
        >>> connection(
        ...   UpdateItem(table, {"h": 0})
        ...     .set("a", ":v")
        ...     .expression_attribute_value("v", 2)
        ...     .return_values_all_new()
        ... ).attributes
        {u'a': 2, u'h': 0, u'b': 2}
        """
        return self.__return_values.all_new()

    @proxy
    def return_values_updated_old(self):
        """
        >>> connection(PutItem(table, {"h": 0, "a": 1, "b": 2}))
        <LowVoltage.actions.put_item.PutItemResponse object at ...>
        >>> connection(
        ...   UpdateItem(table, {"h": 0})
        ...     .set("a", ":v")
        ...     .expression_attribute_value("v", 2)
        ...     .return_values_updated_old()
        ... ).attributes
        {u'a': 1}
        """
        return self.__return_values.updated_old()

    @proxy
    def return_values_updated_new(self):
        """
        >>> connection(PutItem(table, {"h": 0, "a": 1, "b": 2}))
        <LowVoltage.actions.put_item.PutItemResponse object at ...>
        >>> connection(
        ...   UpdateItem(table, {"h": 0})
        ...     .set("a", ":v")
        ...     .expression_attribute_value("v", 2)
        ...     .return_values_updated_new()
        ... ).attributes
        {u'a': 2}
        """
        return self.__return_values.updated_new()

    @proxy
    def return_values_none(self):
        """
        >>> connection(PutItem(table, {"h": 0, "a": 1, "b": 2}))
        <LowVoltage.actions.put_item.PutItemResponse object at ...>
        >>> print connection(
        ...   UpdateItem(table, {"h": 0})
        ...     .set("a", ":v")
        ...     .expression_attribute_value("v", 2)
        ...     .return_values_none()
        ... ).attributes
        None
        """
        return self.__return_values.none()

    @proxy
    def return_consumed_capacity_total(self):
        """
        >>> connection(
        ...   UpdateItem(table, {"h": 4}).set("gh", "h").set("gr", "h")
        ...     .return_consumed_capacity_total()
        ... ).consumed_capacity.capacity_units
        3.0
        """
        return self.__return_consumed_capacity.total()

    @proxy
    def return_consumed_capacity_indexes(self):
        """
        >>> c = connection(
        ...   UpdateItem(table, {"h": 5}).set("gh", "h").set("gr", "h")
        ...     .return_consumed_capacity_indexes()
        ... ).consumed_capacity
        >>> c.capacity_units
        3.0
        >>> c.table.capacity_units
        1.0
        >>> c.global_secondary_indexes["gsi"].capacity_units
        2.0
        """
        return self.__return_consumed_capacity.indexes()

    @proxy
    def return_consumed_capacity_none(self):
        """
        >>> print connection(
        ...   UpdateItem(table, {"h": 6}).set("gh", "h").set("gr", "h")
        ...     .return_consumed_capacity_none()
        ... ).consumed_capacity
        None
        """
        return self.__return_consumed_capacity.none()

    @proxy
    def return_item_collection_metrics_size(self):
        """
        >>> m = connection(
        ...   UpdateItem(table2, {"h": 0, "r1": 0}).set("a", "h")
        ...     .return_item_collection_metrics_size()
        ... ).item_collection_metrics
        >>> m.item_collection_key
        {u'h': 0}
        >>> m.size_estimate_range_gb
        [0.0, 1.0]
        """
        return self.__return_item_collection_metrics.size()

    @proxy
    def return_item_collection_metrics_none(self):
        """
        >>> print connection(
        ...   UpdateItem(table2, {"h": 1, "r1": 0}).set("a", "h")
        ...     .return_item_collection_metrics_none()
        ... ).item_collection_metrics
        None
        """
        return self.__return_item_collection_metrics.none()


class UpdateItemUnitTests(_tst.UnitTests):
    def test_name(self):
        self.assertEqual(UpdateItem("Table", {"hash": 42}).name, "UpdateItem")

    def test_key(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": 42}).payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"N": "42"}},
            }
        )

    def test_set(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": 42}).set("a", ":v").payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"N": "42"}},
                "UpdateExpression": "SET a=:v",
            }
        )

    def test_several_sets(self):
        self.assertIn(
            UpdateItem("Table", {"hash": 42}).set("a", ":v").set("b", ":w").payload,
            [
                {
                    "TableName": "Table",
                    "Key": {"hash": {"N": "42"}},
                    "UpdateExpression": "SET a=:v, b=:w",
                },
                {
                    "TableName": "Table",
                    "Key": {"hash": {"N": "42"}},
                    "UpdateExpression": "SET b=:w, a=:v",
                }
            ]
        )

    def test_remove(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": 42}).remove("a").remove("b").payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"N": "42"}},
                "UpdateExpression": "REMOVE a, b",
            }
        )

    def test_add(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": 42}).add("a", "v").payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"N": "42"}},
                "UpdateExpression": "ADD a :v",
            }
        )

    def test_several_adds(self):
        self.assertIn(
            UpdateItem("Table", {"hash": 42}).add("a", "v").add("b", "w").payload,
            [
                {
                    "TableName": "Table",
                    "Key": {"hash": {"N": "42"}},
                    "UpdateExpression": "ADD a :v, b :w",
                },
                {
                    "TableName": "Table",
                    "Key": {"hash": {"N": "42"}},
                    "UpdateExpression": "ADD b :w, a :v",
                }
            ]
        )

    def test_delete(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": 42}).delete("a", "v").payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"N": "42"}},
                "UpdateExpression": "DELETE a :v",
            }
        )

    def test_several_deletes(self):
        self.assertIn(
            UpdateItem("Table", {"hash": 42}).delete("a", "v").delete("b", "w").payload,
            [
                {
                    "TableName": "Table",
                    "Key": {"hash": {"N": "42"}},
                    "UpdateExpression": "DELETE a :v, b :w",
                },
                {
                    "TableName": "Table",
                    "Key": {"hash": {"N": "42"}},
                    "UpdateExpression": "DELETE b :w, a :v",
                }
            ]
        )

    def test_expression_attribute_value(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": 42}).expression_attribute_value("v", u"value").payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"N": "42"}},
                "ExpressionAttributeValues": {":v": {"S": "value"}},
            }
        )

    def test_expression_attribute_name(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": 42}).expression_attribute_name("n", "path").payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"N": "42"}},
                "ExpressionAttributeNames": {"#n": "path"},
            }
        )

    def test_condition_expression(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": 42}).condition_expression("a=b").payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"N": "42"}},
                "ConditionExpression": "a=b",
            }
        )

    def test_return_values_all_new(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": u"h"}).return_values_all_new().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ReturnValues": "ALL_NEW",
            }
        )

    def test_return_values_all_old(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": u"h"}).return_values_all_old().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ReturnValues": "ALL_OLD",
            }
        )

    def test_return_values_updated_new(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": u"h"}).return_values_updated_new().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ReturnValues": "UPDATED_NEW",
            }
        )

    def test_return_values_updated_old(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": u"h"}).return_values_updated_old().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ReturnValues": "UPDATED_OLD",
            }
        )

    def test_return_values_none(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": u"h"}).return_values_none().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ReturnValues": "NONE",
            }
        )

    def test_return_consumed_capacity_total(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": u"h"}).return_consumed_capacity_total().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ReturnConsumedCapacity": "TOTAL",
            }
        )

    def test_return_consumed_capacity_indexes(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": u"h"}).return_consumed_capacity_indexes().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ReturnConsumedCapacity": "INDEXES",
            }
        )

    def test_return_consumed_capacity_none(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": u"h"}).return_consumed_capacity_none().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ReturnConsumedCapacity": "NONE",
            }
        )

    def test_return_item_collection_metrics_size(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": u"h"}).return_item_collection_metrics_size().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ReturnItemCollectionMetrics": "SIZE",
            }
        )

    def test_return_item_collection_metrics_none(self):
        self.assertEqual(
            UpdateItem("Table", {"hash": u"h"}).return_item_collection_metrics_none().payload,
            {
                "TableName": "Table",
                "Key": {"hash": {"S": "h"}},
                "ReturnItemCollectionMetrics": "NONE",
            }
        )


class UpdateItemLocalIntegTests(_tst.LocalIntegTestsWithTableH):
    def test_set(self):
        r = self.connection(
            _lv.UpdateItem("Aaa", {"h": u"set"})
                .set("a", ":v")
                .set("#p", ":w")
                .expression_attribute_value("v", "aaa")
                .expression_attribute_value("w", "bbb")
                .expression_attribute_name("p", "b")
        )

        self.assertEqual(
            self.connection(_lv.GetItem("Aaa", {"h": u"set"})).item,
            {"h": "set", "a": "aaa", "b": "bbb"}
        )

    def test_complex_update(self):
        self.connection(
            _lv.PutItem(
                "Aaa",
                {
                    "h": u"complex",
                    "a": "a",
                    "b": "b",
                    "c": "c",
                    "d": set([41, 43]),
                    "e": 42,
                    "f": set([41, 42, 43]),
                    "g": set([39, 40]),
                }
            )
        )

        r = self.connection(
            _lv.UpdateItem("Aaa", {"h": u"complex"})
                .set("a", ":s")
                .set("b", ":i")
                .remove("c")
                .add("d", "s")
                .add("e", "i")
                .delete("f", "s")
                .delete("g", "s")
                .expression_attribute_value("s", set([42, 43]))
                .expression_attribute_value("i", 52)
                .return_values_all_new()
        )

        self.assertEqual(
            r.attributes,
            {
                "h": u"complex",
                "a": set([42, 43]),
                "b": 52,
                "d": set([41, 42, 43]),
                "e": 94,
                "f": set([41]),
                "g": set([39, 40]),
            }
        )

    def test_condition_expression(self):
        self.connection(_lv.PutItem("Aaa", {"h": u"expr", "a": 42, "b": 42}))

        r = self.connection(
            _lv.UpdateItem("Aaa", {"h": u"expr"})
                .set("checked", ":true")
                .expression_attribute_value("true", True)
                .condition_expression("a=b")
                .return_values_all_new()
        )

        self.assertEqual(
            r.attributes,
            {"h": u"expr", "a": 42, "b": 42, "checked": True}
        )


class UpdateItemConnectedIntegTests(_tst.ConnectedIntegTestsWithTable):
    def tearDown(self):
        self.connection(_lv.DeleteItem(self.table, self.tab_key))
        super(UpdateItemConnectedIntegTests, self).tearDown()

    def test_return_consumed_capacity_indexes_without_indexed_attribute(self):
        r = self.connection(
            _lv.UpdateItem(self.table, self.tab_key)
                .set("a", ":a").expression_attribute_value("a", "a")
                .return_consumed_capacity_indexes()
        )

        self.assertEqual(r.consumed_capacity.capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity.global_secondary_indexes, None)
        self.assertEqual(r.consumed_capacity.local_secondary_indexes, None)
        self.assertEqual(r.consumed_capacity.table.capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity.table_name, self.table)

    def test_return_consumed_capacity_indexes(self):
        r = self.connection(
            _lv.UpdateItem(self.table, self.tab_key)
                .set("gsi_h", ":gsi_h").expression_attribute_value("gsi_h", u"1")
                .set("gsi_r", ":gsi_r").expression_attribute_value("gsi_r", 1)
                .set("lsi_r", ":lsi_r").expression_attribute_value("lsi_r", 2)
                .return_consumed_capacity_indexes()
        )

        self.assertEqual(r.consumed_capacity.global_secondary_indexes["gsi"].capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity.local_secondary_indexes["lsi"].capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity.table.capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity.capacity_units, 3.0)
        self.assertEqual(r.consumed_capacity.table_name, self.table)

    def test_return_consumed_capacity_indexes_with_locally_indexed_attribute_only(self):
        r = self.connection(
            _lv.UpdateItem(self.table, self.tab_key)
                .set("lsi_r", ":lsi_r").expression_attribute_value("lsi_r", 2)
                .return_consumed_capacity_indexes()
        )

        self.assertEqual(r.consumed_capacity.capacity_units, 2.0)
        self.assertEqual(r.consumed_capacity.global_secondary_indexes, None)
        self.assertEqual(r.consumed_capacity.local_secondary_indexes["lsi"].capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity.table.capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity.table_name, self.table)

    def test_return_consumed_capacity_indexes_with_globally_indexed_attribute_only(self):
        r = self.connection(
            _lv.UpdateItem(self.table, self.tab_key)
                .set("gsi_h", ":gsi_h").expression_attribute_value("gsi_h", u"1")
                .set("gsi_r", ":gsi_r").expression_attribute_value("gsi_r", 1)
                .return_consumed_capacity_indexes()
        )

        self.assertEqual(r.consumed_capacity.capacity_units, 2.0)
        self.assertEqual(r.consumed_capacity.global_secondary_indexes["gsi"].capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity.local_secondary_indexes, None)
        self.assertEqual(r.consumed_capacity.table.capacity_units, 1.0)
        self.assertEqual(r.consumed_capacity.table_name, self.table)

    def test_return_item_collection_metrics_size(self):
        r = self.connection(
            _lv.UpdateItem(self.table, self.tab_key)
                .set("a", ":a").expression_attribute_value("a", "a")
                .return_item_collection_metrics_size()
        )

        self.assertEqual(r.item_collection_metrics.item_collection_key, {"tab_h": u"0"})
        self.assertEqual(r.item_collection_metrics.size_estimate_range_gb[0], 0.0)
        self.assertEqual(r.item_collection_metrics.size_estimate_range_gb[1], 1.0)



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from ...batch_get_item import BatchGetItemLocalIntegTests
from ...batch_write_item import BatchWriteItemLocalIntegTests
from ...create_table import CreateTableLocalIntegTests, CreateTableErrorLocalIntegTests
from ...delete_item import DeleteItemLocalIntegTests
from ...delete_table import DeleteTableLocalIntegTests
from ...describe_table import DescribeTableLocalIntegTests
from ...get_item import GetItemLocalIntegTests
from ...list_tables import ListTablesLocalIntegTests
from ...put_item import PutItemLocalIntegTests
from ...query import QueryLocalIntegTests
from ...scan import ScanLocalIntegTests
from ...update_item import UpdateItemLocalIntegTests
from ...update_table import UpdateTableLocalIntegTests



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from ..list_tables_iterator import ListTablesIteratorUnitTests
from ..batch_get_item_iterator import BatchGetItemIteratorUnitTests
from ..batch_delete_item import BatchDeleteItemUnitTests
from ..batch_put_item import BatchPutItemUnitTests



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import time

import LowVoltage as _lv
import LowVoltage.testing as _tst


def WaitForTableActivation(connection, table):
    """Make "DescribeTable" actions until the table's status is "ACTIVE"."""

    r = connection(_lv.DescribeTable(table))
    while r.table.table_status != "ACTIVE" or (r.table.global_secondary_indexes is not None and any(gsi.index_status != "ACTIVE" for gsi in r.table.global_secondary_indexes)):
        # @todo Use a policy to choose polling interval? Same in WaitForTableDeletion.
        time.sleep(3)
        r = connection(_lv.DescribeTable(table))


class WaitForTableActivationLocalIntegTests(_tst.LocalIntegTests):
    def tearDown(self):
        self.connection(_lv.DeleteTable("Aaa"))
        super(WaitForTableActivationLocalIntegTests, self).tearDown()

    def test(self):
        self.connection(_lv.CreateTable("Aaa").hash_key("h", _lv.STRING).provisioned_throughput(1, 1))
        _lv.WaitForTableActivation(self.connection, "Aaa")
        self.assertEqual(self.connection(_lv.DescribeTable("Aaa")).table.table_status, "ACTIVE")


class WaitForTableActivationConnectedIntegTests(_tst.ConnectedIntegTests):
    def setUp(self):
        super(WaitForTableActivationConnectedIntegTests, self).setUp()
        self.table = self.make_table_name()

    def tearDown(self):
        self.connection(_lv.DeleteTable(self.table))
        super(WaitForTableActivationConnectedIntegTests, self).tearDown()

    def test(self):
        self.connection(
            _lv.CreateTable(self.table).hash_key("tab_h", _lv.STRING).range_key("tab_r", _lv.NUMBER).provisioned_throughput(1, 1)
                .global_secondary_index("gsi").hash_key("gsi_h", _lv.STRING).range_key("gsi_r", _lv.NUMBER).project_all().provisioned_throughput(1, 1)
        )
        _lv.WaitForTableActivation(self.connection, self.table)
        r = self.connection(_lv.DescribeTable(self.table))
        self.assertEqual(r.table.table_status, "ACTIVE")
        self.assertEqual(r.table.global_secondary_indexes[0].index_status, "ACTIVE")



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
@todo Link to user guide (authentication).

Authentication credentials are passed to the connection as a credentials provider.
On each request, the connection retrieves a key/secret pair from the credentials provider, uses it to sign the request, and then discards it.
This allows credentials rotation in the same long-lived connection: the credential provider just has to return the new credentials.

.. py:class:: Credentials

    The interface to be implemented by all credential providers. Note that you must not inherit from this class, just implement the same interface.

    .. py:method:: get()

        Return the key/secret/token to be used to sign the next request. If the credentials are permanent, return ``None`` as token.

        :type: tuple of (string, string, ``None`` or string)
"""

import datetime
import os

import requests

import LowVoltage.testing as _tst
import LowVoltage.exceptions as _exn


class StaticCredentials(object):
    """
    The simplest credential provider: a constant key/secret pair (and optionally a session token if you know what you're doing).
    """

    def __init__(self, key, secret, token=None):
        self.__key = key
        self.__secret = secret
        self.__token = token

    def get(self):
        return self.__key, self.__secret, self.__token


class EnvironmentCredentials(object):
    """
    Credential provider reading the ``AWS_ACCESS_KEY_ID``, ``AWS_SECRET_ACCESS_KEY`` and optionally ``AWS_SECURITY_TOKEN`` environment variables.
    """

    def __init__(self):
        os.environ["AWS_ACCESS_KEY_ID"]
        os.environ["AWS_SECRET_ACCESS_KEY"]

    def get(self):
        return (os.environ["AWS_ACCESS_KEY_ID"], os.environ["AWS_SECRET_ACCESS_KEY"], os.environ.get("AWS_SECURITY_TOKEN"))


class Ec2RoleCredentials(object):
    """
    Credentials provider using EC2 instance metadata to retrieve temporary, automatically rotated, credentials
    from the `IAM role of the instance <http://docs.aws.amazon.com/IAM/latest/UserGuide/roles-usingrole-ec2instance.html>`__.
    Usable *only* on an EC2 instance with an IAM role assigned.

    :param requests_session: a ``Session`` object from the `python-requests <http://python-requests.org>`__ library. Typically not used. Leave it to ``None`` and one will be created for you.
    """

    def __init__(self, requests_session=None):
        if requests_session is None:
            requests_session = requests.Session()

        self.__session = requests_session
        try:
            role = self.__session.get("http://169.254.169.254/latest/meta-data/iam/security-credentials/").text
        except requests.exceptions.RequestException as e:
            raise _exn.NetworkError(e)
        except Exception as e:
            raise _exn.UnknownError(e)
        self.__creds_uri = "http://169.254.169.254/latest/meta-data/iam/security-credentials/{}".format(role)

        self.__key = None
        self.__secret = None
        self.__token = None

        # Dependency injection through monkey-patching
        self.__now = datetime.datetime.utcnow

    def get(self):
        now = self.__now()
        if self.__needs_refresh(now):
            self.__refresh(now)

        return self.__key, self.__secret, self.__token

    def __needs_refresh(self, now):
        if self.__key is None:
            return True
        elif now >= self.__expiration:
            return True
        elif now >= self.__next_refresh:
            return True
        else:
            return False

    def __refresh(self, now):
        try:
            creds = self.__session.get(self.__creds_uri).json()
        except requests.exceptions.RequestException as e:
            raise _exn.NetworkError(e)
        except Exception as e:
            raise _exn.UnknownError(e)
        # {
        #     u'Code': u'Success',
        #     u'LastUpdated': u'2015-04-24T13:36:55Z',
        #     u'AccessKeyId': u'ASIAISBLBZIEQJ6USRAQ',
        #     u'SecretAccessKey': u'xxx',
        #     u'Token': u'yyy',
        #     u'Expiration': u'2015-04-24T19:36:54Z',
        #     u'Type': u'AWS-HMAC'
        # }
        self.__key = creds["AccessKeyId"]
        self.__secret = creds["SecretAccessKey"]
        self.__token = creds["Token"]
        # Refresh every hour and 15 minutes before expiration: http://docs.aws.amazon.com/IAM/latest/UserGuide/roles-usingrole-ec2instance.html
        self.__expiration = datetime.datetime.strptime(creds["Expiration"], "%Y-%m-%dT%H:%M:%SZ") - datetime.timedelta(minutes=15)
        self.__next_refresh = now + datetime.timedelta(hours=1)


class Ec2RoleCredentialsUnitTests(_tst.UnitTestsWithMocks):
    def setUp(self):
        super(Ec2RoleCredentialsUnitTests, self).setUp()
        self.session = self.mocks.create("session")
        self.response = self.mocks.create("response")

    def test_refresh_scenarios(self):
        self.session.expect.get("http://169.254.169.254/latest/meta-data/iam/security-credentials/").andReturn(self.response.object)
        self.response.expect.text.andReturn("RoleName")

        credentials = Ec2RoleCredentials(self.session.object)

        self.now = self.mocks.replace("credentials._Ec2RoleCredentials__now")

        self.now.expect().andReturn(datetime.datetime(2015, 04, 24, 12, 30, 0))
        self.session.expect.get("http://169.254.169.254/latest/meta-data/iam/security-credentials/RoleName").andReturn(self.response.object)
        self.response.expect.json().andReturn({"AccessKeyId": "key1", "SecretAccessKey": "secret1", "Token": "token1", "Expiration": "2015-04-24T15:00:30Z"})

        self.assertEqual(credentials.get(), ("key1", "secret1", "token1"))

        self.now.expect().andReturn(datetime.datetime(2015, 04, 24, 12, 31, 0))
        self.assertEqual(credentials.get(), ("key1", "secret1", "token1"))

        self.now.expect().andReturn(datetime.datetime(2015, 04, 24, 13, 29, 0))
        self.assertEqual(credentials.get(), ("key1", "secret1", "token1"))

        self.now.expect().andReturn(datetime.datetime(2015, 04, 24, 13, 30, 0))
        self.session.expect.get("http://169.254.169.254/latest/meta-data/iam/security-credentials/RoleName").andReturn(self.response.object)
        self.response.expect.json().andReturn({"AccessKeyId": "key2", "SecretAccessKey": "secret2", "Token": "token3", "Expiration": "2015-04-24T14:00:30Z"})
        self.assertEqual(credentials.get(), ("key2", "secret2", "token3"))

        self.now.expect().andReturn(datetime.datetime(2015, 04, 24, 13, 45, 0))
        self.assertEqual(credentials.get(), ("key2", "secret2", "token3"))

        self.now.expect().andReturn(datetime.datetime(2015, 04, 24, 13, 46, 0))
        self.session.expect.get("http://169.254.169.254/latest/meta-data/iam/security-credentials/RoleName").andReturn(self.response.object)
        self.response.expect.json().andReturn({"AccessKeyId": "key3", "SecretAccessKey": "secret3", "Token": "token3", "Expiration": "2015-04-24T18:00:30Z"})
        self.assertEqual(credentials.get(), ("key3", "secret3", "token3"))

    def test_network_error_during_construction(self):
        with self.assertRaises(_exn.NetworkError):
            Ec2RoleCredentials()

    def test_unknown_error_during_construction(self):
        self.session.expect.get("http://169.254.169.254/latest/meta-data/iam/security-credentials/").andRaise(Exception)

        with self.assertRaises(_exn.UnknownError):
            Ec2RoleCredentials(self.session.object)

    def test_network_error_during_refresh(self):
        self.session.expect.get("http://169.254.169.254/latest/meta-data/iam/security-credentials/").andReturn(self.response.object)
        self.response.expect.text.andReturn("RoleName")

        credentials = Ec2RoleCredentials(self.session.object)

        self.now = self.mocks.replace("credentials._Ec2RoleCredentials__now")

        self.now.expect().andReturn(datetime.datetime(2015, 04, 24, 12, 30, 0))
        self.session.expect.get("http://169.254.169.254/latest/meta-data/iam/security-credentials/RoleName").andRaise(requests.exceptions.RequestException)

        with self.assertRaises(_exn.NetworkError):
            credentials.get()

    def test_unknown_error_during_refresh(self):
        self.session.expect.get("http://169.254.169.254/latest/meta-data/iam/security-credentials/").andReturn(self.response.object)
        self.response.expect.text.andReturn("RoleName")

        credentials = Ec2RoleCredentials(self.session.object)

        self.now = self.mocks.replace("credentials._Ec2RoleCredentials__now")

        self.now.expect().andReturn(datetime.datetime(2015, 04, 24, 12, 30, 0))
        self.session.expect.get("http://169.254.169.254/latest/meta-data/iam/security-credentials/RoleName").andRaise(Exception)

        with self.assertRaises(_exn.UnknownError):
            credentials.get()



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import LowVoltage.testing as _tst
from .conversion import _convert_value_to_db, _convert_dict_to_db


def ScalarValue(name):
    class ScalarValue(object):
        def __init__(self, parent):
            self.__value = None
            self.__parent = parent

        def set(self, value):
            self.__value = value
            return self.__parent

        @property
        def payload(self):
            data = {}
            if self.__value is not None:
                data[name] = self.__value
            return data

    return ScalarValue


def DictValue(name):
    class DictValue(object):
        def __init__(self, parent):
            self.__data = {}
            self.__parent = parent

        def add(self, name, value):
            self.__data[name] = value
            return self.__parent

        @property
        def payload(self):
            data = {}
            if len(self.__data) != 0:
                data[name] = self.__data
            return data

    return DictValue


def CommaSeparatedStringsValue(name):
    class CommaSeparatedStringsValue(object):
        def __init__(self, parent):
            self.__values = []
            self.__parent = parent

        def add(self, *values):
            for value in values:
                if isinstance(value, basestring):
                    value = [value]
                self.__values.extend(value)
            return self.__parent

        @property
        def payload(self):
            data = {}
            if len(self.__values) != 0:
                data[name] = ", ".join(self.__values)
            return data

    return CommaSeparatedStringsValue


class ConditionExpression(ScalarValue("ConditionExpression")):
    def set(self, expression):
        """
        Set the ConditionExpression, making the request conditional.
        It will raise a :exc:`.ConditionalCheckFailedException` if the condition is not met.
        """
        return super(ConditionExpression, self).set(expression)


class ConsistentRead(ScalarValue("ConsistentRead")):
    def false(self):
        """
        Set ConsistentRead to False.
        The request will use eventually consistent reads.
        """
        return self.set(False)

    def true(self):
        """
        Set ConsistentRead to True.
        The request will use strong consistent reads.
        """
        return self.set(True)


class ExclusiveStartKey(ScalarValue("ExclusiveStartKey")):
    def set(self, key):
        """
        Set ExclusiveStartKey. The request will only scan items that are after this key.
        This is typically the :attr:`~{}Response.last_evaluated_key` of a previous response.
        """
        return super(ExclusiveStartKey, self).set(_convert_dict_to_db(key))


class ExpressionAttributeNames(DictValue("ExpressionAttributeNames")):
    def add(self, synonym, name):
        """
        Add a synonym for an attribute name to ExpressionAttributeNames.
        Useful for attributes whose names don't play well with ProjectionExpression, ConditionExpression or UpdateExpression because they contain a dot or brackets.
        """
        return super(ExpressionAttributeNames, self).add("#" + synonym, name)


class ExpressionAttributeValues(DictValue("ExpressionAttributeValues")):
    def add(self, name, value):
        """
        Add a named value to ExpressionAttributeValues.
        """
        return super(ExpressionAttributeValues, self).add(":" + name, _convert_value_to_db(value))


class FilterExpression(ScalarValue("FilterExpression")):
    def set(self, expression):
        """
        Set the FilterExpression. The response will contain only items that match.
        """
        return super(FilterExpression, self).set(expression)


class Limit(ScalarValue("Limit")):
    def set(self, limit):
        """
        Set Limit. The request will scan at most this number of items.
        """
        return super(Limit, self).set(limit)


class ProjectionExpression(CommaSeparatedStringsValue("ProjectionExpression")):
    def add(self, *names):
        """
        Add name(s) to ProjectionExpression.
        The request will return only projected attributes.

        Note that this method accepts a variable number of names or iterables.  @todo A section in the user guide describing those methods.
        """
        return super(ProjectionExpression, self).add(*names)


class ReturnConsumedCapacity(ScalarValue("ReturnConsumedCapacity")):
    def indexes(self):
        """
        Set ReturnConsumedCapacity to INDEXES.
        The response will contain the capacity consumed by this request detailled on the table and the indexes.
        """
        return self.set("INDEXES")

    def none(self):
        """
        Set ReturnConsumedCapacity to NONE.
        The response will not contain the capacity consumed by this request.
        """
        return self.set("NONE")

    def total(self):
        """
        Set ReturnConsumedCapacity to TOTAL.
        The response will contain the total capacity consumed by this request.
        """
        return self.set("TOTAL")


class ReturnItemCollectionMetrics(ScalarValue("ReturnItemCollectionMetrics")):
    def none(self):
        """
        Set ReturnItemCollectionMetrics to NONE.
        The response will not contain any item collection metrics.
        """
        return self.set("NONE")

    def size(self):
        """
        Set ReturnItemCollectionMetrics to SIZE.
        If the table has a local secondary index, the response will contain metrics about the size of item collections that were touched.
        """
        return self.set("SIZE")


class ReturnValues(ScalarValue("ReturnValues")):
    def all_new(self):
        """
        Set ReturnValues to ALL_NEW.
        The response will contain all the attributes of the item in its new state.
        """
        return self.set("ALL_NEW")

    def all_old(self):
        """
        Set ReturnValues to ALL_OLD.
        The response will contain all the attributes of the item in its previous state.
        """
        return self.set("ALL_OLD")

    def none(self):
        """
        Set ReturnValues to NONE.
        The response will not include the attributes of the item.
        """
        return self.set("NONE")

    def updated_new(self):
        """
        Set ReturnValues to UPDATED_NEW.
        The response will contain the just-updated attributes of the item in its new state.
        """
        return self.set("UPDATED_NEW")

    def updated_old(self):
        """
        Set ReturnValues to UPDATED_OLD.
        The response will contain the just-updated attributes of the item in its previous state.
        """
        return self.set("UPDATED_OLD")


class Select(ScalarValue("Select")):
    def all_attributes(self):
        """
        Set Select to ALL_ATTRIBUTES. The response will contain all attributes of the matching items.
        """
        return self.set("ALL_ATTRIBUTES")

    def all_projected_attributes(self):
        """
        Set Select to ALL_PROJECTED_ATTRIBUTES. Usable only when querying an index.
        The response will contain the attributes of the matching items that are projected on the index.
        """
        return self.set("ALL_PROJECTED_ATTRIBUTES")

    def count(self):
        """
        Set Select to COUNT. The response will contain only the count of matching items.
        """
        return self.set("COUNT")


def proxy(*proxy_args):
    bases = {
        "condition_expression": ConditionExpression.set,
        "consistent_read_false": ConsistentRead.false,
        "consistent_read_true": ConsistentRead.true,
        "exclusive_start_key": ExclusiveStartKey.set,
        "expression_attribute_name": ExpressionAttributeNames.add,
        "expression_attribute_value": ExpressionAttributeValues.add,
        "filter_expression": FilterExpression.set,
        "limit": Limit.set,
        "project": ProjectionExpression.add,
        "return_consumed_capacity_indexes": ReturnConsumedCapacity.indexes,
        "return_consumed_capacity_none": ReturnConsumedCapacity.none,
        "return_consumed_capacity_total": ReturnConsumedCapacity.total,
        "return_item_collection_metrics_none": ReturnItemCollectionMetrics.none,
        "return_item_collection_metrics_size": ReturnItemCollectionMetrics.size,
        "return_values_all_new": ReturnValues.all_new,
        "return_values_all_old": ReturnValues.all_old,
        "return_values_none": ReturnValues.none,
        "return_values_updated_new": ReturnValues.updated_new,
        "return_values_updated_old": ReturnValues.updated_old,
        "select_all_attributes": Select.all_attributes,
        "select_all_projected_attributes": Select.all_projected_attributes,
        "select_count": Select.count,
    }

    if len(proxy_args) != 1 or isinstance(proxy_args[0], basestring):
        def decorator(method):
            method.__doc__ = bases[method.__name__].__doc__.format(*proxy_args) + "\n" + method.__doc__
            return method
        return decorator
    else:
        method, = proxy_args
        method.__doc__ = bases[method.__name__].__doc__.format() + "\n" + method.__doc__
        return method



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import datetime
import glob
import io
import os
import stat
import subprocess
import tarfile

import requests
try:
    from testresources import TestResourceManager, ResourcedTestCase
except ImportError:  # pragma no cover (Test code)
    class TestResourceManager(object):
        pass

    class ResourcedTestCase(object):
        pass

import LowVoltage as _lv


class DynamoDbLocalResourceManager(TestResourceManager):
    def make(self, dependencies):
        self.__download_if_needed()

        self.__process = subprocess.Popen(
            # ["sleep 7; java -Djava.library.path=./DynamoDBLocal_lib -jar DynamoDBLocal.jar -inMemory -port 65432"],
            ["java", "-Djava.library.path=./DynamoDBLocal_lib", "-jar", "DynamoDBLocal.jar", "-inMemory", "-port", "65432"],
            cwd=".dynamodblocal",
            stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        return _lv.Connection(
            "us-west-2",
            _lv.StaticCredentials("DummyKey", "DummySecret"),
            endpoint="http://localhost:65432/",
            retry_policy=_lv.ExponentialBackoffRetryPolicy(1, 2, 5),
        )

    def __download_if_needed(self):  # pragma no cover (Test code)
        if not os.path.exists(".dynamodblocal/DynamoDBLocal.jar"):
            archive = requests.get("http://dynamodb-local.s3-website-us-west-2.amazonaws.com/dynamodb_local_latest").content
            tarfile.open(fileobj=io.BytesIO(archive)).extractall(".dynamodblocal")
            # Fix permissions, needed at least when running with Cygwin's Python
            for f in glob.glob(".dynamodblocal/DynamoDBLocal_lib/*"):
                os.chmod(f, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)

    def clean(self, resource):
        self.__process.kill()


class LocalIntegTests(ResourcedTestCase):
    resources = [("connection", DynamoDbLocalResourceManager())]

    before_start = datetime.datetime.utcnow()
    after_end = before_start + datetime.timedelta(minutes=10)

    def assertDateTimeIsReasonable(self, t):
        self.assertGreaterEqual(t, self.before_start)
        self.assertLessEqual(t, self.after_end)


class LocalIntegTestsWithTableH(LocalIntegTests):
    def setUp(self):
        super(LocalIntegTestsWithTableH, self).setUp()
        self.connection(
            _lv.CreateTable("Aaa").hash_key("h", _lv.STRING).provisioned_throughput(1, 1)
        )

    def tearDown(self):
        self.connection(_lv.DeleteTable("Aaa"))
        super(LocalIntegTestsWithTableH, self).tearDown()


class LocalIntegTestsWithTableHR(LocalIntegTests):
    def setUp(self):
        super(LocalIntegTestsWithTableHR, self).setUp()
        self.connection(
            _lv.CreateTable("Aaa")
                .hash_key("h", _lv.STRING)
                .range_key("r", _lv.NUMBER)
                .provisioned_throughput(1, 1)
        )

    def tearDown(self):
        self.connection(_lv.DeleteTable("Aaa"))
        super(LocalIntegTestsWithTableHR, self).tearDown()



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import base64
import numbers
import sys

import LowVoltage.testing as _tst


def _convert_dict_to_db(attributes):
    return {
        key: _convert_value_to_db(val)
        for key, val in attributes.iteritems()
    }


def _convert_value_to_db(value):
    if isinstance(value, unicode):
        return {"S": value}
    elif isinstance(value, bytes):
        return {"B": base64.b64encode(value).decode("utf8")}
    elif isinstance(value, bool):
        return {"BOOL": value}
    elif isinstance(value, numbers.Integral):
        return {"N": str(value)}
    elif value is None:
        return {"NULL": True}
    elif isinstance(value, (set, frozenset)):
        if len(value) == 0:
            raise TypeError
        elif all(isinstance(v, numbers.Integral) for v in value):
            return {"NS": [str(n) for n in value]}
        elif all(isinstance(v, unicode) for v in value):
            return {"SS": [s for s in value]}
        elif all(isinstance(v, bytes) for v in value):
            return {"BS": [base64.b64encode(b).decode("utf8") for b in value]}
        else:
            raise TypeError
    elif isinstance(value, list):
        return {"L": [_convert_value_to_db(v) for v in value]}
    elif isinstance(value, dict):
        return {"M": {n: _convert_value_to_db(v) for n, v in value.iteritems()}}
    else:
        raise TypeError


def _convert_db_to_dict(attributes):
    return {
        key: _convert_db_to_value(val)
        for key, val in attributes.iteritems()
    }


def _convert_db_to_value(value):
    if "S" in value:
        return value["S"]
    elif "B" in value:
        return bytes(base64.b64decode(value["B"].encode("utf8")))
    elif "BOOL" in value:
        return value["BOOL"]
    elif "N" in value:
        return int(value["N"])
    elif "NULL" in value:
        return None
    elif "NS" in value:
        return set(int(v) for v in value["NS"])
    elif "SS" in value:
        return set(value["SS"])
    elif "BS" in value:
        return set(bytes(base64.b64decode(v.encode("utf8"))) for v in value["BS"])
    elif "L" in value:
        return [_convert_db_to_value(v) for v in value["L"]]
    elif "M" in value:
        return {n: _convert_db_to_value(v) for n, v in value["M"].iteritems()}
    else:
        raise TypeError


class ConversionUnitTests(_tst.UnitTests):
    def test_convert_unicode_value_to_db(self):
        self.assertEqual(_convert_value_to_db(u"éoà"), {"S": u"éoà"})

    def test_convert_bytes_value_to_db(self):
        self.assertEqual(_convert_value_to_db(b"\xFF\x00\xAB"), {"B": u"/wCr"})

    def test_convert_bool_value_to_db(self):
        self.assertEqual(_convert_value_to_db(True), {"BOOL": True})
        self.assertEqual(_convert_value_to_db(False), {"BOOL": False})

    def test_convert_int_value_to_db(self):
        self.assertEqual(_convert_value_to_db(42), {"N": "42"})

    def test_convert_none_value_to_db(self):
        self.assertEqual(_convert_value_to_db(None), {"NULL": True})

    def test_convert_set_of_int_value_to_db(self):
        self.assertIn(_convert_value_to_db(set([42, 43])), [{"NS": ["42", "43"]}, {"NS": ["43", "42"]}])

    def test_convert_frozenset_value_to_db(self):
        self.assertIn(_convert_value_to_db(frozenset([42, 43])), [{"NS": ["42", "43"]}, {"NS": ["43", "42"]}])

    def test_convert_set_of_unicode_value_to_db(self):
        self.assertIn(_convert_value_to_db(set([u"éoà", u"bar"])), [{"SS": [u"éoà", u"bar"]}, {"SS": [u"bar", u"éoà"]}])

    def test_convert_set_of_byte_value_to_db(self):
        self.assertIn(_convert_value_to_db(set([b"\xFF\x00\xAB", b"bar"])), [{"BS": [u"/wCr", u"YmFy"]}, {"BS": [u"YmFy", u"/wCr"]}])

    def test_convert_list_value_to_db(self):
        self.assertEqual(_convert_value_to_db([True, 42]), {"L": [{"BOOL": True}, {"N": "42"}]})

    def test_convert_dict_value_to_db(self):
        self.assertEqual(_convert_value_to_db({"a": True, "b": 42}), {"M": {"a": {"BOOL": True}, "b": {"N": "42"}}})

    def test_convert_empty_set_value_to_db(self):
        with self.assertRaises(TypeError):
            _convert_value_to_db(set())

    def test_convert_set_of_tuple_value_to_db(self):
        with self.assertRaises(TypeError):
            _convert_value_to_db(set([(1, 2)]))

    def test_convert_heterogenous_set_value_to_db(self):
        with self.assertRaises(TypeError):
            _convert_value_to_db(set([1, "2"]))

    def test_convert_tuple_value_to_db(self):
        with self.assertRaises(TypeError):
            _convert_value_to_db((1, 2))

    def test_convert_db_to_unicode_value(self):
        self.assertEqual(_convert_db_to_value({"S": u"éoà"}), u"éoà")

    def test_convert_db_to_bytes_value(self):
        self.assertEqual(_convert_db_to_value({"B": u"/wCr"}), b"\xFF\x00\xAB")

    def test_convert_db_to_bool_value(self):
        self.assertEqual(_convert_db_to_value({"BOOL": True}), True)
        self.assertEqual(_convert_db_to_value({"BOOL": False}), False)

    def test_convert_db_to_int_value(self):
        self.assertEqual(_convert_db_to_value({"N": "42"}), 42)

    def test_convert_db_to_none_value(self):
        self.assertEqual(_convert_db_to_value({"NULL": True}), None)

    def test_convert_db_to_set_of_int_value(self):
        self.assertEqual(_convert_db_to_value({"NS": ["42", "43"]}), set([42, 43]))

    def test_convert_db_to_set_of_unicode_value(self):
        self.assertEqual(_convert_db_to_value({"SS": [u"éoà", u"bar"]}), set([u"éoà", u"bar"]))

    def test_convert_db_to_set_of_byte_value(self):
        self.assertEqual(_convert_db_to_value({"BS": [u"/wCr", u"YmFy"]}), set([b"\xFF\x00\xAB", b"bar"]))

    def test_convert_db_to_list_value(self):
        self.assertEqual(_convert_db_to_value({"L": [{"BOOL": True}, {"N": "42"}]}), [True, 42])

    def test_convert_db_to_dict_value(self):
        self.assertEqual(_convert_db_to_value({"M": {"a": {"BOOL": True}, "b": {"N": "42"}}}), {"a": True, "b": 42})

    def test_convert_db_to_empty_dict_value(self):
        with self.assertRaises(TypeError):
            _convert_db_to_value({})

    def test_convert_db_to_empty_list_value(self):
        with self.assertRaises(TypeError):
            _convert_db_to_value([])

    def test_convert_db_to_string_value(self):
        with self.assertRaises(TypeError):
            _convert_db_to_value("SSS")

    def test_convert_empty_dict_to_db(self):
        self.assertEqual(_convert_dict_to_db({}), {})

    def test_convert_dict_to_db(self):
        self.assertEqual(_convert_dict_to_db({"a": 42}), {"a": {"N": "42"}})

    def test_convert_db_to_empty_dict(self):
        self.assertEqual(_convert_db_to_dict({}), {})

    def test_convert_db_to_dict(self):
        self.assertEqual(_convert_db_to_dict({"a": {"N": "42"}}), {"a": 42})



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import LowVoltage.testing as _tst


class _Boolean(object):
    def __and__(self, other):
        return _BooleanExpression(self, "AND", other)

    def __or__(self, other):
        return _BooleanExpression(self, "OR", other)

    def __invert__(self):
        return _BooleanNegation(self)


class _BooleanExpression(_Boolean):
    def __init__(self, left, operator, right):
        if not isinstance(left, _Boolean):
            raise TypeError
        if not isinstance(operator, basestring):
            raise TypeError
        if not isinstance(right, _Boolean):
            raise TypeError
        self.__left = left
        self.__operator = operator
        self.__right = right

    def bool(self):
        return "({}) {} ({})".format(self.__left.bool(), self.__operator, self.__right.bool())


class _BooleanNegation(_Boolean):
    def __init__(self, operand):
        if not isinstance(operand, _Boolean):
            raise TypeError
        self.__operand = operand

    def bool(self):
        return "NOT ({})".format(self.__operand.bool())


class _ComparisonExpression(_Boolean):
    def __init__(self, left, operator, right):
        if not isinstance(left, _Atom):
            raise TypeError
        if not isinstance(operator, basestring):
            raise TypeError
        if not isinstance(right, _Atom):
            raise TypeError
        self.__left = left
        self.__operator = operator
        self.__right = right

    def bool(self):
        return "{}{}{}".format(self.__left.atom(), self.__operator, self.__right.atom())


class _Atom(object):
    def __eq__(self, other):
        return _ComparisonExpression(self, "=", other)

    def __ne__(self, other):
        return _ComparisonExpression(self, "<>", other)

    def __lt__(self, other):
        return _ComparisonExpression(self, "<", other)

    def __le__(self, other):
        return _ComparisonExpression(self, "<=", other)

    def __gt__(self, other):
        return _ComparisonExpression(self, ">", other)

    def __ge__(self, other):
        return _ComparisonExpression(self, ">=", other)


class Attr(_Atom):
    def __init__(self, name):
        if not isinstance(name, basestring):
            raise TypeError
        self.__name = name

    def atom(self):
        return self.__name


class Val(_Atom):
    def __init__(self, label):
        if not isinstance(label, basestring):
            raise TypeError
        self.__label = label

    def atom(self):
        return ":{}".format(self.__label)


class In(_Boolean):
    def __init__(self, elem, set):
        if not isinstance(elem, _Atom):
            raise TypeError
        if not all(isinstance(elem, _Atom) for elem in set):
            raise TypeError
        self.__elem = elem
        self.__set = set

    def bool(self):
        return "{} IN ({})".format(self.__elem.atom(), ", ".join(elem.atom() for elem in self.__set))


class Between(_Boolean):
    def __init__(self, elem, low, high):
        if not isinstance(elem, _Atom):
            raise TypeError
        if not isinstance(low, _Atom):
            raise TypeError
        if not isinstance(high, _Atom):
            raise TypeError
        self.__elem = elem
        self.__low = low
        self.__high = high

    def bool(self):
        return "{} BETWEEN {} AND {}".format(self.__elem.atom(), self.__low.atom(), self.__high.atom())


class AttributeExists(_Boolean):
    def __init__(self, name):
        if not isinstance(name, basestring):
            raise TypeError
        self.__name = name

    def bool(self):
        return "attribute_exists({})".format(self.__name)


class Contains(_Boolean):
    def __init__(self, left, right):
        if not isinstance(left, _Atom):
            raise TypeError
        if not isinstance(right, _Atom):
            raise TypeError
        self.__left = left
        self.__right = right

    def bool(self):
        return "contains({}, {})".format(self.__left.atom(), self.__right.atom())


class BeginsWith(_Boolean):
    def __init__(self, left, right):
        if not isinstance(left, _Atom):
            raise TypeError
        if not isinstance(right, _Atom):
            raise TypeError
        self.__left = left
        self.__right = right

    def bool(self):
        return "begins_with({}, {})".format(self.__left.atom(), self.__right.atom())


class ConditionExpressionUnitTests(_tst.UnitTests):
    def test_atoms_comparison(self):
        self.assertEqual((Attr("a") == Attr("b")).bool(), "a=b")
        self.assertEqual((Attr("a") == Val("b")).bool(), "a=:b")
        self.assertEqual((Val("a") == Attr("b")).bool(), ":a=b")
        self.assertEqual((Val("a") == Val("b")).bool(), ":a=:b")

    def test_comparisons(self):
        self.assertEqual((Attr("a") == Attr("b")).bool(), "a=b")
        self.assertEqual((Attr("a") != Attr("b")).bool(), "a<>b")
        self.assertEqual((Attr("a") < Attr("b")).bool(), "a<b")
        self.assertEqual((Attr("a") <= Attr("b")).bool(), "a<=b")
        self.assertEqual((Attr("a") > Attr("b")).bool(), "a>b")
        self.assertEqual((Attr("a") >= Attr("b")).bool(), "a>=b")

    def test_boolean_algebra(self):
        # Too many parentheses are needed because of lower priority of bitwise operators ("&", "|" and "~"),
        # but we cannot override keywords ("and", "or" and "not") that would have higher priority.
        self.assertEqual(((Attr("a") == Attr("b")) & (Attr("c") == Attr("d"))).bool(), "(a=b) AND (c=d)")
        self.assertEqual(((Attr("a") == Attr("b")) | (Attr("c") == Attr("d"))).bool(), "(a=b) OR (c=d)")
        self.assertEqual((~((Attr("a") == Attr("b")) | (Attr("c") == Attr("d")))).bool(), "NOT ((a=b) OR (c=d))")

    def test_functions(self):
        # Erf, we cannot redefine list.__contains__ so we cannot obtain the syntax:
        # Attr("a") in [Val("b"), Val("c")]
        # And anyway the result of __contains__ is converted to a boolean by "in".
        # So we cannot implement Attr("a") in Set(Val("b"), Val("c")) either.
        self.assertEqual(In(Attr("a"), [Val("b"), Val("c")]).bool(), "a IN (:b, :c)")
        self.assertEqual(Between(Attr("a"), Val("b"), Val("c")).bool(), "a BETWEEN :b AND :c")
        self.assertEqual(AttributeExists("a").bool(), "attribute_exists(a)")
        self.assertEqual(Contains(Val("a"), Attr("b")).bool(), "contains(:a, b)")
        self.assertEqual(BeginsWith(Val("a"), Attr("b")).bool(), "begins_with(:a, b)")

    def test_missing_parentheses_in_boolean_algebra(self):
        # But at least, missing parentheses are caught early
        with self.assertRaises(TypeError):
            Attr("a") == Attr("b") & Attr("c") == Attr("d")
        with self.assertRaises(TypeError):
            (Attr("a") == Attr("b")) & Attr("c") == Attr("d")
        with self.assertRaises(TypeError):
            Attr("a") == Attr("b") & (Attr("c") == Attr("d"))

    def test_plain_wrong_types(self):
        with self.assertRaises(TypeError):
            Attr("a") & (Attr("c") == Attr("d"))
        with self.assertRaises(TypeError):
            (Attr("a") == Attr("c")) & Attr("d")
        with self.assertRaises(TypeError):
            (Attr("a") == Attr("c")) == Attr("d")
        with self.assertRaises(TypeError):
            In("a", [Attr("a")])
        with self.assertRaises(TypeError):
            In(Attr("a"), ["a"])
        with self.assertRaises(TypeError):
            _BooleanExpression(Attr("a"), "==", Attr("a") == Attr("b"))
        with self.assertRaises(TypeError):
            _BooleanExpression(Attr("a") == Attr("b"), 42, Attr("a") == Attr("b"))
        with self.assertRaises(TypeError):
            _BooleanNegation(Attr("a"))
        with self.assertRaises(TypeError):
            _ComparisonExpression("a", "==", Attr("b"))
        with self.assertRaises(TypeError):
            _ComparisonExpression(Attr("a"), 42, Attr("b"))
        with self.assertRaises(TypeError):
            _ComparisonExpression(Attr("a"), "==", "b")
        with self.assertRaises(TypeError):
            Attr(42)
        with self.assertRaises(TypeError):
            Val(42)
        with self.assertRaises(TypeError):
            Between("a", Attr("b"), Attr("c"))
        with self.assertRaises(TypeError):
            Between(Attr("a"), "b", Attr("c"))
        with self.assertRaises(TypeError):
            Between(Attr("a"), Attr("b"), "c")
        with self.assertRaises(TypeError):
            AttributeExists(Attr("a"))
        with self.assertRaises(TypeError):
            Contains("a", Attr("b"))
        with self.assertRaises(TypeError):
            Contains(Attr("a"), "b")
        with self.assertRaises(TypeError):
            BeginsWith("a", Attr("b"))
        with self.assertRaises(TypeError):
            BeginsWith(Attr("a"), "b")



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from ...batch_delete_item import BatchDeleteItemLocalIntegTests
from ...batch_get_item_iterator import BatchGetItemIteratorLocalIntegTests
from ...batch_put_item import BatchPutItemLocalIntegTests
from ...list_tables_iterator import ListTablesIteratorLocalIntegTests
from ...query_iterator import QueryIteratorLocalIntegTests
from ...scan_iterator import ScanIteratorLocalIntegTests
from ...wait_for_table_activation import WaitForTableActivationLocalIntegTests
from ...wait_for_table_deletion import WaitForTableDeletionLocalIntegTests



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
@todo Link to user guide (error management).
"""


class Error(Exception):
    """
    The base class of all exceptions raised by the package.
    """
    retryable = False


class UnknownError(Error):
    """
    Exception raised for errors that can't be attributed to the client, the network or the server.

    You should ``except Error`` instead of this one.
    """


class ServerError(Error):
    """
    Exception raised when the problem can be blamed on the server.
    Typically DynamoDB returned a 5XX status code or a response that couldn't be decoded.
    """
    retryable = True


class NetworkError(Error):
    """
    Exception raised when the problem can be blamed on the network.
    Connection refused, timeout, etc.
    """
    retryable = True


class ClientError(Error):
    """
    Exception raised when the problem can be blamed on the client.

    See bellow for specialized exceptions for some client errors.
    """


class UnknownClientError(ClientError):
    """
    Exception raised for errors that can be attributed to the client but are not known by the package.
    Typically DynamoDB returned a 4XX status code but we couldn't match the returned type to any known client error.

    You should ``except ClientError`` instead of this one.

    Feel free to `open an issue <https://github.com/jacquev6/LowVoltage/issues/new?title=UnknownClientError>`__
    if your ``except ClientError`` clause catches an :exc:`UnknownClientError`.
    Put the exception details and we'll add it to the package.
    """


class BuilderError(ClientError):
    """
    Exception raised when you make a mistake while using a builder.

    For example if you call :meth:`.CreateTable.project` when you don't have an active index.
    Or :meth:`.BatchGetItem.keys` when you don't have an active table.
    """


# All 4XXs from http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html
# and "Errors" sections of all actions (like http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_GetItem.html#API_GetItem_Errors)


class ConditionalCheckFailedException(ClientError):
    """
    Raised by conditional updates when the condition is not met.
    See :meth:`.PutItem.condition_expression`, :meth:`.UpdateItem.condition_expression` and :meth:`.DeleteItem.condition_expression`.
    """


class IncompleteSignature(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """


class InvalidAction(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """


class InvalidClientTokenId(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """


class InvalidParameterCombination(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """


class InvalidParameterValue(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """


class InvalidQueryParameter(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """


class ItemCollectionSizeLimitExceededException(ClientError):
    """
    Exception raised when the item collection of a hash key is too large (in a table with a LSI).
    """


class LimitExceededException(ClientError):
    """
    Exception raised when too many tables are beeing modified at the same time.
    Or when there are too many tables.

    See `the reference of the CreateTable action <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_CreateTable.html#API_CreateTable_Errors>`__.
    """


class MalformedQueryString(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """


class MissingAction(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """


class MissingAuthenticationToken(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """


class MissingParameter(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """


class OptInRequired(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """


class ProvisionedThroughputExceededException(ClientError):
    """
    Exception raised when the provisioned throughput is reached.
    """
    retryable = True


class RequestExpired(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """
    retryable = True


class ResourceInUseException(ClientError):
    """
    Exception raise when trying to modify a table that's not in the "ACTIVE" state.
    """
    retryable = True


class ResourceNotFoundException(ClientError):
    """
    Exception raised when trying to use a non-existent table.
    """


class Throttling(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """
    retryable = True


class ValidationError(ClientError):
    """
    See `common errors <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    """


class ValidationException(ClientError):
    """
    Exception raised when the request is invalid.
    """


# Error discovered by chance, not documented


class AccessDeniedException(ClientError):
    """
    Exception `not documented <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    Seems to be raised when credentials are valid, but the operation is not allowed by IAM policies.
    """


class InvalidSignatureException(ClientError):
    """
    Exception `not documented <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    Seems to be raised when credentials are not valid.
    """


class SerializationException(ClientError):
    """
    Exception `not documented <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    Seems to be raised when a number argument (passed as a string) cannot be converted to an actual number.
    """


class UnrecognizedClientException(ClientError):
    """
    Exception `not documented <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/CommonErrors.html>`__.
    Seems to be raised when using (valid) temporary credentials but an invalid token.
    """


# Sorted by decreasing suffix length to ensure BasicConnection._raise finds the right class using str.endswith.
client_errors = sorted(
    [
        ("ConditionalCheckFailedException", ConditionalCheckFailedException),
        ("IncompleteSignature", IncompleteSignature),
        ("InvalidAction", InvalidAction),
        ("InvalidClientTokenId", InvalidClientTokenId),
        ("InvalidParameterCombination", InvalidParameterCombination),
        ("InvalidParameterValue", InvalidParameterValue),
        ("InvalidQueryParameter", InvalidQueryParameter),
        ("ItemCollectionSizeLimitExceededException", ItemCollectionSizeLimitExceededException),
        ("LimitExceededException", LimitExceededException),
        ("MalformedQueryString", MalformedQueryString),
        ("MissingAction", MissingAction),
        ("MissingAuthenticationToken", MissingAuthenticationToken),
        ("MissingParameter", MissingParameter),
        ("OptInRequired", OptInRequired),
        ("ProvisionedThroughputExceededException", ProvisionedThroughputExceededException),
        ("RequestExpired", RequestExpired),
        ("ResourceInUseException", ResourceInUseException),
        ("ResourceNotFoundException", ResourceNotFoundException),
        ("Throttling", Throttling),
        ("ValidationError", ValidationError),
        ("ValidationException", ValidationException),

        ("AccessDeniedException", AccessDeniedException),
        ("InvalidSignatureException", InvalidSignatureException),
        ("SerializationException", SerializationException),
        ("UnrecognizedClientException", UnrecognizedClientException),
    ],
    key=lambda (prefix, cls): -len(prefix)
)



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
When given a :class:`ListTables`, the connection will return a :class:`ListTablesResponse`:

>>> r = connection(ListTables())
>>> r
<LowVoltage.actions.list_tables.ListTablesResponse ...>
>>> r.table_names
[u'LowVoltage.Tests.Doc.1', u'LowVoltage.Tests.Doc.2']
"""

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .action import Action
from .return_types import _is_str, _is_list_of_str


class ListTablesResponse(object):
    """
    The `ListTables response <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_ListTables.html#API_ListTables_ResponseElements>`__.
    """

    def __init__(
        self,
        LastEvaluatedTableName=None,
        TableNames=None,
        **dummy
    ):
        self.__last_evaluated_table_name = LastEvaluatedTableName
        self.__table_names = TableNames

    @property
    def last_evaluated_table_name(self):
        """
        The name of the last table that was considered during the request.
        If not None, you should give it to :meth:`~ListTables.exclusive_start_table_name` in a subsequent :class:`ListTables`.

        :type: ``None`` or string
        """
        if _is_str(self.__last_evaluated_table_name):  # pragma no branch (Defensive code)
            return self.__last_evaluated_table_name

    @property
    def table_names(self):
        """
        The names of the tables.

        :type: ``None`` or list of string
        """
        if _is_list_of_str(self.__table_names):  # pragma no branch (Defensive code)
            return self.__table_names


class ListTables(Action):
    """
    The `ListTables request <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_ListTables.html#API_ListTables_RequestParameters>`__.
    """

    def __init__(self):
        super(ListTables, self).__init__("ListTables", ListTablesResponse)
        self.__limit = None
        self.__exclusive_start_table_name = None

    @property
    def payload(self):
        data = {}
        if self.__limit is not None:
            data["Limit"] = self.__limit
        if self.__exclusive_start_table_name:
            data["ExclusiveStartTableName"] = self.__exclusive_start_table_name
        return data

    def limit(self, limit):
        """
        Set Limit. The response will contain at most this number of table names.

        >>> r = connection(ListTables().limit(1))
        >>> r.table_names
        [u'LowVoltage.Tests.Doc.1']
        >>> r.last_evaluated_table_name
        u'LowVoltage.Tests.Doc.1'
        """
        self.__limit = limit
        return self

    def exclusive_start_table_name(self, table_name):
        """
        Set ExclusiveStartTableName. The response will contains tables that are after this one.
        Typically the :attr:`~ListTablesResponse.last_evaluated_table_name` of a previous response.

        >>> connection(
        ...   ListTables()
        ...     .exclusive_start_table_name("LowVoltage.Tests.Doc.1")
        ... ).table_names
        [u'LowVoltage.Tests.Doc.2']
        """
        self.__exclusive_start_table_name = table_name
        return self


class ListTablesUnitTests(_tst.UnitTests):
    def test_name(self):
        self.assertEqual(ListTables().name, "ListTables")

    def test_no_arguments(self):
        self.assertEqual(ListTables().payload, {})

    def test_limit(self):
        self.assertEqual(ListTables().limit(42).payload, {"Limit": 42})

    def test_exclusive_start_table_name(self):
        self.assertEqual(ListTables().exclusive_start_table_name("Bar").payload, {"ExclusiveStartTableName": "Bar"})


class ListTablesLocalIntegTests(_tst.LocalIntegTests):
    def setUp(self):
        super(ListTablesLocalIntegTests, self).setUp()
        self.connection(
            _lv.CreateTable("Aaa").hash_key("h", _lv.STRING).provisioned_throughput(1, 2)
        )
        self.connection(
            _lv.CreateTable("Bbb").hash_key("h", _lv.STRING).provisioned_throughput(1, 2)
        )
        self.connection(
            _lv.CreateTable("Ccc").hash_key("h", _lv.STRING).provisioned_throughput(1, 2)
        )

    def tearDown(self):
        self.connection(_lv.DeleteTable("Aaa"))
        self.connection(_lv.DeleteTable("Bbb"))
        self.connection(_lv.DeleteTable("Ccc"))
        super(ListTablesLocalIntegTests, self).tearDown()

    def test_all_arguments(self):
        r = self.connection(_lv.ListTables().exclusive_start_table_name("Aaa").limit(1))

        self.assertEqual(r.last_evaluated_table_name, "Bbb")
        self.assertEqual(r.table_names[0], "Bbb")

    def test_no_arguments(self):
        r = self.connection(_lv.ListTables())

        self.assertEqual(r.table_names[0], "Aaa")
        self.assertEqual(r.table_names[1], "Bbb")
        self.assertEqual(r.table_names[2], "Ccc")



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from ...batch_get_item import BatchGetItemConnectedIntegTests
from ...batch_write_item import BatchWriteItemConnectedIntegTests
from ...delete_item import DeleteItemConnectedIntegTests
from ...get_item import GetItemConnectedIntegTests
from ...put_item import PutItemConnectedIntegTests
from ...query import QueryConnectedIntegTests
from ...scan import ScanConnectedIntegTests
from ...update_item import UpdateItemConnectedIntegTests



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import datetime

from LowVoltage.actions.conversion import _convert_dict_to_db, _convert_value_to_db, _convert_db_to_dict, _convert_db_to_value


def _is_dict(d):
    return isinstance(d, dict)


def _is_str(s):
    return isinstance(s, basestring)


def _is_float(s):
    return isinstance(s, float)


def _is_int(s):
    return isinstance(s, (int, long))


def _is_list_of_dict(l):
    return isinstance(l, list) and all(_is_dict(e) for e in l)


def _is_list_of_str(l):
    return isinstance(l, list) and all(_is_str(e) for e in l)


def _is_list_of_float(l):
    return isinstance(l, list) and all(_is_float(e) for e in l)


class TableDescription(object):
    """
    `TableDescription <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_TableDescription.html>`__.
    """

    def __init__(
        self,
        AttributeDefinitions=None,
        CreationDateTime=None,
        GlobalSecondaryIndexes=None,
        ItemCount=None,
        KeySchema=None,
        LocalSecondaryIndexes=None,
        ProvisionedThroughput=None,
        TableName=None,
        TableSizeBytes=None,
        TableStatus=None,
        **dummy
    ):
        self.__attribute_definitions = AttributeDefinitions
        self.__creation_date_time = CreationDateTime
        self.__global_secondary_indexes = GlobalSecondaryIndexes
        self.__item_count = ItemCount
        self.__key_schema = KeySchema
        self.__local_secondary_indexes = LocalSecondaryIndexes
        self.__provisioned_throughput = ProvisionedThroughput
        self.__table_name = TableName
        self.__table_size_bytes = TableSizeBytes
        self.__table_status = TableStatus

    @property
    def attribute_definitions(self):
        """
        :type: ``None`` or list of :class:`.AttributeDefinition`
        """
        if _is_list_of_dict(self.__attribute_definitions):  # pragma no branch (Defensive code)
            return [AttributeDefinition(**d) for d in self.__attribute_definitions]

    @property
    def creation_date_time(self):
        """
        :type: ``None`` or :class:`~datetime.datetime`
        """
        if _is_float(self.__creation_date_time):  # pragma no branch (Defensive code)
            return datetime.datetime.utcfromtimestamp(self.__creation_date_time)

    @property
    def global_secondary_indexes(self):
        """
        :type: ``None`` or list of :class:`.GlobalSecondaryIndexDescription`
        """
        if _is_list_of_dict(self.__global_secondary_indexes):  # pragma no branch (Defensive code)
            return [GlobalSecondaryIndexDescription(**d) for d in self.__global_secondary_indexes]

    @property
    def item_count(self):
        """
        :type: ``None`` or long
        """
        if _is_int(self.__item_count):  # pragma no branch (Defensive code)
            return long(self.__item_count)

    @property
    def key_schema(self):
        """
        :type: ``None`` or list of :class:`.KeySchemaElement`
        """
        if _is_list_of_dict(self.__key_schema):  # pragma no branch (Defensive code)
            return [KeySchemaElement(**e) for e in self.__key_schema]

    @property
    def local_secondary_indexes(self):
        """
        :type: ``None`` or list of :class:`.LocalSecondaryIndexDescription`
        """
        if _is_list_of_dict(self.__local_secondary_indexes):  # pragma no branch (Defensive code)
            return [LocalSecondaryIndexDescription(**d) for d in self.__local_secondary_indexes]

    @property
    def provisioned_throughput(self):
        """
        :type: ``None`` or :class:`.ProvisionedThroughputDescription`
        """
        if _is_dict(self.__provisioned_throughput):  # pragma no branch (Defensive code)
            return ProvisionedThroughputDescription(**self.__provisioned_throughput)

    @property
    def table_name(self):
        """
        :type: ``None`` or string
        """
        if _is_str(self.__table_name):  # pragma no branch (Defensive code)
            return self.__table_name

    @property
    def table_size_bytes(self):
        """
        :type: ``None`` or long
        """
        if _is_int(self.__table_size_bytes):  # pragma no branch (Defensive code)
            return long(self.__table_size_bytes)

    @property
    def table_status(self):
        """
        :type: ``None`` or string
        """
        if _is_str(self.__table_status):  # pragma no branch (Defensive code)
            return self.__table_status


class AttributeDefinition(object):
    """
    `AttributeDefinition <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_AttributeDefinition.html>`__.
    """

    def __init__(
        self,
        AttributeName=None,
        AttributeType=None,
        **dummy
    ):
        self.__attribute_name = AttributeName
        self.__attribute_type = AttributeType

    @property
    def attribute_name(self):
        """
        :type: ``None`` or string
        """
        if _is_str(self.__attribute_name):  # pragma no branch (Defensive code)
            return self.__attribute_name

    @property
    def attribute_type(self):
        """
        :type: ``None`` or string
        """
        if _is_str(self.__attribute_type):  # pragma no branch (Defensive code)
            return self.__attribute_type


class GlobalSecondaryIndexDescription(object):
    """
    `GlobalSecondaryIndexDescription <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_GlobalSecondaryIndexDescription.html>`__.
    """

    def __init__(
        self,
        IndexName=None,
        IndexSizeBytes=None,
        IndexStatus=None,
        ItemCount=None,
        KeySchema=None,
        Projection=None,
        ProvisionedThroughput=None,
        **dummy
    ):
        self.__index_name = IndexName
        self.__index_size_bytes = IndexSizeBytes
        self.__index_status = IndexStatus
        self.__item_count = ItemCount
        self.__key_schema = KeySchema
        self.__projection = Projection
        self.__provisioned_throughput = ProvisionedThroughput

    @property
    def index_name(self):
        """
        :type: ``None`` or string
        """
        if _is_str(self.__index_name):  # pragma no branch (Defensive code)
            return self.__index_name

    @property
    def index_size_bytes(self):
        """
        :type: ``None`` or long
        """
        if _is_int(self.__index_size_bytes):  # pragma no branch (Defensive code)
            return long(self.__index_size_bytes)

    @property
    def index_status(self):
        """
        :type: ``None`` or string
        """
        if _is_str(self.__index_status):  # pragma no branch (Defensive code)
            return self.__index_status

    @property
    def item_count(self):
        """
        :type: ``None`` or long
        """
        if _is_int(self.__item_count):  # pragma no branch (Defensive code)
            return long(self.__item_count)

    @property
    def key_schema(self):
        """
        :type: ``None`` or list of :class:`.KeySchemaElement`
        """
        if _is_list_of_dict(self.__key_schema):  # pragma no branch (Defensive code)
            return [KeySchemaElement(**e) for e in self.__key_schema]

    @property
    def projection(self):
        """
        :type: ``None`` or :class:`.Projection`
        """
        if _is_dict(self.__projection):  # pragma no branch (Defensive code)
            return Projection(**self.__projection)

    @property
    def provisioned_throughput(self):
        """
        :type: ``None`` or :class:`.ProvisionedThroughputDescription`
        """
        if _is_dict(self.__provisioned_throughput):  # pragma no branch (Defensive code)
            return ProvisionedThroughputDescription(**self.__provisioned_throughput)


class Projection(object):
    """
    `Projection <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_Projection.html>`__.
    """

    def __init__(
        self,
        NonKeyAttributes=None,
        ProjectionType=None,
        **dummy
    ):
        self.__non_key_attributes = NonKeyAttributes
        self.__projection_type = ProjectionType

    @property
    def non_key_attributes(self):
        """
        :type: ``None`` or list of string
        """
        if _is_list_of_str(self.__non_key_attributes):  # pragma no branch (Defensive code)
            return self.__non_key_attributes

    @property
    def projection_type(self):
        """
        :type: ``None`` or string
        """
        if _is_str(self.__projection_type):  # pragma no branch (Defensive code)
            return self.__projection_type


class ProvisionedThroughputDescription(object):
    """
    `ProvisionedThroughputDescription <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_ProvisionedThroughputDescription.html>`__.
    """

    def __init__(
        self,
        LastDecreaseDateTime=None,
        LastIncreaseDateTime=None,
        NumberOfDecreasesToday=None,
        ReadCapacityUnits=None,
        WriteCapacityUnits=None,
        **dummy
    ):
        self.__last_decrease_date_time = LastDecreaseDateTime
        self.__last_increase_date_time = LastIncreaseDateTime
        self.__number_of_decreases_today = NumberOfDecreasesToday
        self.__read_capacity_units = ReadCapacityUnits
        self.__write_capacity_units = WriteCapacityUnits

    @property
    def last_decrease_date_time(self):
        """
        :type: ``None`` or :class:`~datetime.datetime`
        """
        if _is_float(self.__last_decrease_date_time):  # pragma no branch (Defensive code)
            return datetime.datetime.utcfromtimestamp(self.__last_decrease_date_time)

    @property
    def last_increase_date_time(self):
        """
        :type: ``None`` or :class:`~datetime.datetime`
        """
        if _is_float(self.__last_increase_date_time):  # pragma no branch (Defensive code)
            return datetime.datetime.utcfromtimestamp(self.__last_increase_date_time)

    @property
    def number_of_decreases_today(self):
        """
        :type: ``None`` or long
        """
        if _is_int(self.__number_of_decreases_today):  # pragma no branch (Defensive code)
            return long(self.__number_of_decreases_today)

    @property
    def read_capacity_units(self):
        """
        :type: ``None`` or long
        """
        if _is_int(self.__read_capacity_units):  # pragma no branch (Defensive code)
            return long(self.__read_capacity_units)

    @property
    def write_capacity_units(self):
        """
        :type: ``None`` or long
        """
        if _is_int(self.__write_capacity_units):  # pragma no branch (Defensive code)
            return long(self.__write_capacity_units)


class KeySchemaElement(object):
    """
    `KeySchemaElement <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_KeySchemaElement.html>`__.
    """

    def __init__(
        self,
        AttributeName=None,
        KeyType=None,
        **dummy
    ):
        self.__attribute_name = AttributeName
        self.__key_type = KeyType

    @property
    def attribute_name(self):
        """
        :type: ``None`` or string
        """
        if _is_str(self.__attribute_name):  # pragma no branch (Defensive code)
            return self.__attribute_name

    @property
    def key_type(self):
        """
        :type: ``None`` or string
        """
        if _is_str(self.__key_type):  # pragma no branch (Defensive code)
            return self.__key_type


class LocalSecondaryIndexDescription(object):
    """
    `LocalSecondaryIndexDescription <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_LocalSecondaryIndexDescription.html>`__.
    """

    def __init__(
        self,
        IndexName=None,
        IndexSizeBytes=None,
        ItemCount=None,
        KeySchema=None,
        Projection=None,
        **dummy
    ):
        self.__index_name = IndexName
        self.__index_size_bytes = IndexSizeBytes
        self.__item_count = ItemCount
        self.__key_schema = KeySchema
        self.__projection = Projection

    @property
    def index_name(self):
        """
        :type: ``None`` or string
        """
        if _is_str(self.__index_name):  # pragma no branch (Defensive code)
            return self.__index_name

    @property
    def index_size_bytes(self):
        """
        :type: ``None`` or long
        """
        if _is_int(self.__index_size_bytes):  # pragma no branch (Defensive code)
            return long(self.__index_size_bytes)

    @property
    def item_count(self):
        """
        :type: ``None`` or long
        """
        if _is_int(self.__item_count):  # pragma no branch (Defensive code)
            return long(self.__item_count)

    @property
    def key_schema(self):
        """
        :type: ``None`` or list of :class:`.KeySchemaElement`
        """
        if _is_list_of_dict(self.__key_schema):  # pragma no branch (Defensive code)
            return [KeySchemaElement(**e) for e in self.__key_schema]

    @property
    def projection(self):
        """
        :type: ``None`` or :class:`.Projection`
        """
        if _is_dict(self.__projection):  # pragma no branch (Defensive code)
            return Projection(**self.__projection)


class ConsumedCapacity(object):
    """
    `ConsumedCapacity <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_ConsumedCapacity.html>`__.
    """

    def __init__(
        self,
        CapacityUnits=None,
        GlobalSecondaryIndexes=None,
        LocalSecondaryIndexes=None,
        Table=None,
        TableName=None,
        **dummy
    ):
        self.__capacity_units = CapacityUnits
        self.__global_secondary_indexes = GlobalSecondaryIndexes
        self.__local_secondary_indexes = LocalSecondaryIndexes
        self.__table = Table
        self.__table_name = TableName

    @property
    def capacity_units(self):
        """
        The total capacity units consumed by the request.

        :type: ``None`` or float
        """
        if _is_float(self.__capacity_units):  # pragma no branch (Defensive code)
            return float(self.__capacity_units)

    @property
    def global_secondary_indexes(self):
        """
        The capacity consumed on GSIs.

        :type: ``None`` or dict of string (index name) to :class:`.Capacity`
        """
        if _is_dict(self.__global_secondary_indexes):  # pragma no branch (Defensive code)
            return {n: Capacity(**v) for n, v in self.__global_secondary_indexes.iteritems()}

    @property
    def local_secondary_indexes(self):
        """
        The capacity consumed on LSIs.

        :type: ``None`` or dict of string (index name) to :class:`.Capacity`
        """
        if _is_dict(self.__local_secondary_indexes):  # pragma no branch (Defensive code)
            return {n: Capacity(**v) for n, v in self.__local_secondary_indexes.iteritems()}

    @property
    def table(self):
        """
        The capacity consumed on the table itself.

        :type: ``None`` or :class:`.Capacity`
        """
        if _is_dict(self.__table):  # pragma no branch (Defensive code)
            return Capacity(**self.__table)

    @property
    def table_name(self):
        """
        The name of the table.

        :type: ``None`` or string
        """
        if _is_str(self.__table_name):  # pragma no branch (Defensive code)
            return self.__table_name


class Capacity(object):
    """
    `Capacity <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_Capacity.html>`__.
    """

    def __init__(
        self,
        CapacityUnits=None,
        **dummy
    ):
        self.__capacity_units = CapacityUnits

    @property
    def capacity_units(self):
        """
        Actual units of consumed capacity.

        :type: ``None`` or float
        """
        if _is_float(self.__capacity_units):  # pragma no branch (Defensive code)
            return float(self.__capacity_units)


class ItemCollectionMetrics(object):
    """
    `ItemCollectionMetrics <http://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_ItemCollectionMetrics.html>`__.
    """

    def __init__(
        self,
        ItemCollectionKey=None,
        SizeEstimateRangeGB=None,
        **dummy
    ):
        self.__item_collection_key = ItemCollectionKey
        self.__size_estimate_range_gb = SizeEstimateRangeGB

    @property
    def item_collection_key(self):
        """
        Hash key of the collection whose size is estimated.

        :type: ``None`` or dict
        """
        if _is_dict(self.__item_collection_key):  # pragma no branch (Defensive code)
            return _convert_db_to_dict(self.__item_collection_key)

    @property
    def size_estimate_range_gb(self):
        """
        Range of sizes of the collection in GB.

        :type: ``None`` or list of two float
        """
        if _is_list_of_float(self.__size_estimate_range_gb):  # pragma no branch (Defensive code)
            return [float(e) for e in self.__size_estimate_range_gb]



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>


class Action(object):
    def __init__(self, name, response_class):
        self.name = name
        self.response_class = response_class



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import copy

import LowVoltage as _lv
import LowVoltage.testing as _tst
from .iterator import Iterator


class ScanIterator(Iterator):
    """Make as many "Scan" actions as needed to iterate over all items."""

    @classmethod
    def parallelize(cls, connection, scan, threads):
        """Create several ScanIterator to be used in parallel"""

        return [
            cls(connection, copy.deepcopy(scan).segment(i, threads))
            for i in range(threads)
        ]

    def __init__(self, connection, scan):
        Iterator.__init__(self, connection, scan)

    def process(self, action, r):
        if r.last_evaluated_key is None:
            action = None
        else:
            action.exclusive_start_key(r.last_evaluated_key)
        return action, r.items


class ScanIteratorLocalIntegTests(_tst.LocalIntegTestsWithTableH):
    keys = [u"{:03}".format(k) for k in range(15)]

    def setUp(self):
        super(ScanIteratorLocalIntegTests, self).setUp()
        self.connection(
            _lv.BatchWriteItem().table("Aaa").put(
                {"h": h, "xs": "x" * 300000}  # 300kB items ensure a single Query will return at most 4 items
                for h in self.keys
            )
        )

    def test_simple_scan(self):
        self.assertEqual(
            sorted(item["h"] for item in _lv.ScanIterator(self.connection, _lv.Scan("Aaa"))),
            self.keys
        )

    def test_parallel_scan(self):
        keys = []
        for segment in ScanIterator.parallelize(self.connection, _lv.Scan("Aaa"), 3):
            keys.extend(item["h"] for item in segment)
        self.assertEqual(sorted(keys), self.keys)



# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import unittest

try:
    import MockMockMock
    has_mock = True
except ImportError:  # pragma no cover (Test code)
    has_mock = False


class UnitTests(unittest.TestCase):
    pass


if has_mock:
    class UnitTestsWithMocks(UnitTests):
        class ActionChecker(object):
            def __init__(self, expected_name, expected_payload):
                self.__expected_name = expected_name
                self.__expected_payload = expected_payload

            def __call__(self, args, kwds):
                assert len(args) == 1
                assert len(kwds) == 0
                action, = args
                return action.name == self.__expected_name and action.payload == self.__expected_payload

        def setUp(self):
            super(UnitTestsWithMocks, self).setUp()
            self.mocks = MockMockMock.Engine()

        def tearDown(self):
            self.mocks.tearDown()
            super(UnitTestsWithMocks, self).tearDown()
else:
    class UnitTestsWithMocks(object):  # pragma no cover (Test code)
        pass
